# ruff: noqa: D100
import typing


def vectorized(func: typing.Callable[..., typing.Any]) -> typing.Callable[..., typing.Any]:
    """Decorate a function with annotated function parameters.

    This allows DuckDB to infer that the function should be provided with pyarrow arrays and should expect
    pyarrow array(s) as output.
    """
    import types
    from inspect import signature

    new_func = types.FunctionType(func.__code__, func.__globals__, func.__name__, func.__defaults__, func.__closure__)
    # Construct the annotations:
    import pyarrow as pa

    new_annotations = {}
    sig = signature(func)
    for param in sig.parameters:
        new_annotations[param] = pa.lib.ChunkedArray

    new_func.__annotations__ = new_annotations
    return new_func



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Low-level ADBC bindings for the DuckDB driver."""

import enum
import functools
import importlib.util

import adbc_driver_manager


class StatementOptions(enum.Enum):
    """Statement options specific to the DuckDB driver."""

    #: The number of rows per batch. Defaults to 2048.
    BATCH_ROWS = "adbc.duckdb.query.batch_rows"


def connect(path: str | None = None) -> adbc_driver_manager.AdbcDatabase:
    """Create a low level ADBC connection to DuckDB."""
    if path is None:
        return adbc_driver_manager.AdbcDatabase(driver=driver_path(), entrypoint="duckdb_adbc_init")
    return adbc_driver_manager.AdbcDatabase(driver=driver_path(), entrypoint="duckdb_adbc_init", path=path)


@functools.cache
def driver_path() -> str:
    """Get the path to the DuckDB ADBC driver."""
    duckdb_module_spec = importlib.util.find_spec("_duckdb")
    if duckdb_module_spec is None:
        msg = "Could not find duckdb shared library. Did you pip install duckdb?"
        raise ImportError(msg)
    return duckdb_module_spec.origin



from typing import NamedTuple

from .session import SparkSession


class Database(NamedTuple):  # noqa: D101
    name: str
    description: str | None
    locationUri: str


class Table(NamedTuple):  # noqa: D101
    name: str
    database: str | None
    description: str | None
    tableType: str
    isTemporary: bool


class Column(NamedTuple):  # noqa: D101
    name: str
    description: str | None
    dataType: str
    nullable: bool
    isPartition: bool
    isBucket: bool


class Function(NamedTuple):  # noqa: D101
    name: str
    description: str | None
    className: str
    isTemporary: bool


class Catalog:  # noqa: D101
    def __init__(self, session: SparkSession) -> None:  # noqa: D107
        self._session = session

    def listDatabases(self) -> list[Database]:  # noqa: D102
        res = self._session.conn.sql("select database_name from duckdb_databases()").fetchall()

        def transform_to_database(x: list[str]) -> Database:
            return Database(name=x[0], description=None, locationUri="")

        databases = [transform_to_database(x) for x in res]
        return databases

    def listTables(self) -> list[Table]:  # noqa: D102
        res = self._session.conn.sql("select table_name, database_name, sql, temporary from duckdb_tables()").fetchall()

        def transform_to_table(x: list[str]) -> Table:
            return Table(name=x[0], database=x[1], description=x[2], tableType="", isTemporary=x[3])

        tables = [transform_to_table(x) for x in res]
        return tables

    def listColumns(self, tableName: str, dbName: str | None = None) -> list[Column]:  # noqa: D102
        query = f"""
			select column_name, data_type, is_nullable from duckdb_columns() where table_name = '{tableName}'
		"""
        if dbName:
            query += f" and database_name = '{dbName}'"
        res = self._session.conn.sql(query).fetchall()

        def transform_to_column(x: list[str | bool]) -> Column:
            return Column(name=x[0], description=None, dataType=x[1], nullable=x[2], isPartition=False, isBucket=False)

        columns = [transform_to_column(x) for x in res]
        return columns

    def listFunctions(self, dbName: str | None = None) -> list[Function]:  # noqa: D102
        raise NotImplementedError

    def setCurrentDatabase(self, dbName: str) -> None:  # noqa: D102
        raise NotImplementedError


__all__ = ["Catalog", "Column", "Database", "Function", "Table"]



import duckdb
import typing
from ._typing import StrIntoPyType, PyTypeIds

__all__: list[str] = [
    "BIGINT",
    "BIT",
    "BLOB",
    "BOOLEAN",
    "DATE",
    "DOUBLE",
    "FLOAT",
    "HUGEINT",
    "INTEGER",
    "INTERVAL",
    "SMALLINT",
    "SQLNULL",
    "TIME",
    "TIMESTAMP",
    "TIMESTAMP_MS",
    "TIMESTAMP_NS",
    "TIMESTAMP_S",
    "TIMESTAMP_TZ",
    "TIME_TZ",
    "TINYINT",
    "UBIGINT",
    "UHUGEINT",
    "UINTEGER",
    "USMALLINT",
    "UTINYINT",
    "UUID",
    "VARCHAR",
    "VARIANT",
    "DuckDBPyType",
]

class DuckDBPyType:
    def __eq__(self, other: object) -> bool: ...
    def __getattr__(self, name: str) -> DuckDBPyType: ...
    def __getitem__(self, name: str) -> DuckDBPyType: ...
    def __hash__(self) -> int: ...
    @typing.overload
    def __init__(self, type_str: StrIntoPyType, connection: duckdb.DuckDBPyConnection) -> None: ...
    @typing.overload
    def __init__(self, obj: object) -> None: ...
    @property
    def children(self) -> list[tuple[str, DuckDBPyType | int | list[str]]]: ...
    @property
    def id(self) -> PyTypeIds: ...

BIGINT: DuckDBPyType  # value = BIGINT
BIT: DuckDBPyType  # value = BIT
BLOB: DuckDBPyType  # value = BLOB
BOOLEAN: DuckDBPyType  # value = BOOLEAN
DATE: DuckDBPyType  # value = DATE
DOUBLE: DuckDBPyType  # value = DOUBLE
FLOAT: DuckDBPyType  # value = FLOAT
HUGEINT: DuckDBPyType  # value = HUGEINT
INTEGER: DuckDBPyType  # value = INTEGER
INTERVAL: DuckDBPyType  # value = INTERVAL
SMALLINT: DuckDBPyType  # value = SMALLINT
SQLNULL: DuckDBPyType  # value = "NULL"
TIME: DuckDBPyType  # value = TIME
TIME_NS: DuckDBPyType  # value = TIME_NS
TIMESTAMP: DuckDBPyType  # value = TIMESTAMP
TIMESTAMP_MS: DuckDBPyType  # value = TIMESTAMP_MS
TIMESTAMP_NS: DuckDBPyType  # value = TIMESTAMP_NS
TIMESTAMP_S: DuckDBPyType  # value = TIMESTAMP_S
TIMESTAMP_TZ: DuckDBPyType  # value = TIMESTAMP WITH TIME ZONE
TIME_TZ: DuckDBPyType  # value = TIME WITH TIME ZONE
TINYINT: DuckDBPyType  # value = TINYINT
UBIGINT: DuckDBPyType  # value = UBIGINT
UHUGEINT: DuckDBPyType  # value = UHUGEINT
UINTEGER: DuckDBPyType  # value = UINTEGER
USMALLINT: DuckDBPyType  # value = USMALLINT
UTINYINT: DuckDBPyType  # value = UTINYINT
UUID: DuckDBPyType  # value = UUID
VARCHAR: DuckDBPyType  # value = VARCHAR
VARIANT: DuckDBPyType  # value = VARIANT



"""StringIO buffer wrapper.

BSD 3-Clause License

Copyright (c) 2008-2011, AQR Capital Management, LLC, Lambda Foundry, Inc. and PyData Development Team
All rights reserved.

Copyright (c) 2011-2022, Open source contributors.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from io import StringIO, TextIOBase
from typing import Any


class BytesIOWrapper:
    """Wrapper that wraps a StringIO buffer and reads bytes from it.

    Created for compat with pyarrow read_csv.
    """

    def __init__(self, buffer: StringIO | TextIOBase, encoding: str = "utf-8") -> None:  # noqa: D107
        self.buffer = buffer
        self.encoding = encoding
        # Because a character can be represented by more than 1 byte,
        # it is possible that reading will produce more bytes than n
        # We store the extra bytes in this overflow variable, and append the
        # overflow to the front of the bytestring the next time reading is performed
        self.overflow = b""

    def __getattr__(self, attr: str) -> Any:  # noqa: D105, ANN401
        return getattr(self.buffer, attr)

    def read(self, n: int | None = -1) -> bytes:  # noqa: D102
        assert self.buffer is not None
        bytestring = self.buffer.read(n).encode(self.encoding)
        # When n=-1/n greater than remaining bytes: Read entire file/rest of file
        combined_bytestring = self.overflow + bytestring
        if n is None or n < 0 or n >= len(combined_bytestring):
            self.overflow = b""
            return combined_bytestring
        else:
            to_return = combined_bytestring[:n]
            self.overflow = combined_bytestring[n:]
            return to_return



from .catalog import Catalog  # noqa: D104
from .conf import RuntimeConfig
from .dataframe import DataFrame
from .readwriter import DataFrameWriter
from .session import SparkSession

__all__ = ["Catalog", "DataFrame", "DataFrameWriter", "RuntimeConfig", "SparkSession"]



import duckdb
from duckdb import DuckDBPyConnection
from duckdb.experimental.spark.conf import SparkConf
from duckdb.experimental.spark.exception import ContributionsAcceptedError


class SparkContext:  # noqa: D101
    def __init__(self, master: str) -> None:  # noqa: D107
        self._connection = duckdb.connect(":memory:")
        # This aligns the null ordering with Spark.
        self._connection.execute("set default_null_order='nulls_first_on_asc_last_on_desc'")

    @property
    def connection(self) -> DuckDBPyConnection:  # noqa: D102
        return self._connection

    def stop(self) -> None:  # noqa: D102
        self._connection.close()

    @classmethod
    def getOrCreate(cls, conf: SparkConf | None = None) -> "SparkContext":  # noqa: D102
        raise ContributionsAcceptedError

    @classmethod
    def setSystemProperty(cls, key: str, value: str) -> None:  # noqa: D102
        raise ContributionsAcceptedError

    @property
    def applicationId(self) -> str:  # noqa: D102
        raise ContributionsAcceptedError

    @property
    def defaultMinPartitions(self) -> int:  # noqa: D102
        raise ContributionsAcceptedError

    @property
    def defaultParallelism(self) -> int:  # noqa: D102
        raise ContributionsAcceptedError

    # @property
    # def resources(self) -> Dict[str, ResourceInformation]:
    # 	raise ContributionsAcceptedError

    @property
    def startTime(self) -> str:  # noqa: D102
        raise ContributionsAcceptedError

    @property
    def uiWebUrl(self) -> str:  # noqa: D102
        raise ContributionsAcceptedError

    @property
    def version(self) -> str:  # noqa: D102
        raise ContributionsAcceptedError

    def __repr__(self) -> str:  # noqa: D105
        raise ContributionsAcceptedError

    # def accumulator(self, value: ~T, accum_param: Optional[ForwardRef('AccumulatorParam[T]')] = None
    #     ) -> 'Accumulator[T]':
    # 	pass

    def addArchive(self, path: str) -> None:  # noqa: D102
        raise ContributionsAcceptedError

    def addFile(self, path: str, recursive: bool = False) -> None:  # noqa: D102
        raise ContributionsAcceptedError

    def addPyFile(self, path: str) -> None:  # noqa: D102
        raise ContributionsAcceptedError

    # def binaryFiles(self, path: str, minPartitions: Optional[int] = None
    #     ) -> duckdb.experimental.spark.rdd.RDD[typing.Tuple[str, bytes]]:
    # 	pass

    # def binaryRecords(self, path: str, recordLength: int) -> duckdb.experimental.spark.rdd.RDD[bytes]:
    # 	pass

    # def broadcast(self, value: ~T) -> 'Broadcast[T]':
    # 	pass

    def cancelAllJobs(self) -> None:  # noqa: D102
        raise ContributionsAcceptedError

    def cancelJobGroup(self, groupId: str) -> None:  # noqa: D102
        raise ContributionsAcceptedError

    def dump_profiles(self, path: str) -> None:  # noqa: D102
        raise ContributionsAcceptedError

    # def emptyRDD(self) -> duckdb.experimental.spark.rdd.RDD[typing.Any]:
    # 	pass

    def getCheckpointDir(self) -> str | None:  # noqa: D102
        raise ContributionsAcceptedError

    def getConf(self) -> SparkConf:  # noqa: D102
        raise ContributionsAcceptedError

    def getLocalProperty(self, key: str) -> str | None:  # noqa: D102
        raise ContributionsAcceptedError

    # def hadoopFile(self, path: str, inputFormatClass: str, keyClass: str, valueClass: str,
    #     keyConverter: Optional[str] = None, valueConverter: Optional[str] = None,
    #     conf: Optional[Dict[str, str]] = None, batchSize: int = 0) -> pyspark.rdd.RDD[typing.Tuple[~T, ~U]]:
    # 	pass

    # def hadoopRDD(self, inputFormatClass: str, keyClass: str, valueClass: str, keyConverter: Optional[str] = None,
    #     valueConverter: Optional[str] = None, conf: Optional[Dict[str, str]] = None, batchSize: int = 0
    #     ) -> pyspark.rdd.RDD[typing.Tuple[~T, ~U]]:
    # 	pass

    # def newAPIHadoopFile(self, path: str, inputFormatClass: str, keyClass: str, valueClass: str,
    #     keyConverter: Optional[str] = None, valueConverter: Optional[str] = None,
    #     conf: Optional[Dict[str, str]] = None, batchSize: int = 0) -> pyspark.rdd.RDD[typing.Tuple[~T, ~U]]:
    # 	pass

    # def newAPIHadoopRDD(self, inputFormatClass: str, keyClass: str, valueClass: str,
    #     keyConverter: Optional[str] = None, valueConverter: Optional[str] = None,
    #     conf: Optional[Dict[str, str]] = None, batchSize: int = 0) -> pyspark.rdd.RDD[typing.Tuple[~T, ~U]]:
    # 	pass

    # def parallelize(self, c: Iterable[~T], numSlices: Optional[int] = None) -> pyspark.rdd.RDD[~T]:
    # 	pass

    # def pickleFile(self, name: str, minPartitions: Optional[int] = None) -> pyspark.rdd.RDD[typing.Any]:
    # 	pass

    # def range(self, start: int, end: Optional[int] = None, step: int = 1, numSlices: Optional[int] = None
    #     ) -> pyspark.rdd.RDD[int]:
    # 	pass

    # def runJob(self, rdd: pyspark.rdd.RDD[~T], partitionFunc: Callable[[Iterable[~T]], Iterable[~U]],
    #     partitions: Optional[Sequence[int]] = None, allowLocal: bool = False) -> List[~U]:
    # 	pass

    # def sequenceFile(self, path: str, keyClass: Optional[str] = None, valueClass: Optional[str] = None,
    #     keyConverter: Optional[str] = None, valueConverter: Optional[str] = None, minSplits: Optional[int] = None,
    #     batchSize: int = 0) -> pyspark.rdd.RDD[typing.Tuple[~T, ~U]]:
    # 	pass

    def setCheckpointDir(self, dirName: str) -> None:  # noqa: D102
        raise ContributionsAcceptedError

    def setJobDescription(self, value: str) -> None:  # noqa: D102
        raise ContributionsAcceptedError

    def setJobGroup(self, groupId: str, description: str, interruptOnCancel: bool = False) -> None:  # noqa: D102
        raise ContributionsAcceptedError

    def setLocalProperty(self, key: str, value: str) -> None:  # noqa: D102
        raise ContributionsAcceptedError

    def setLogLevel(self, logLevel: str) -> None:  # noqa: D102
        raise ContributionsAcceptedError

    def show_profiles(self) -> None:  # noqa: D102
        raise ContributionsAcceptedError

    def sparkUser(self) -> str:  # noqa: D102
        raise ContributionsAcceptedError

    # def statusTracker(self) -> duckdb.experimental.spark.status.StatusTracker:
    # 	raise ContributionsAcceptedError

    # def textFile(self, name: str, minPartitions: Optional[int] = None, use_unicode: bool = True
    #     ) -> pyspark.rdd.RDD[str]:
    # 	pass

    # def union(self, rdds: List[pyspark.rdd.RDD[~T]]) -> pyspark.rdd.RDD[~T]:
    # 	pass

    # def wholeTextFiles(self, path: str, minPartitions: Optional[int] = None, use_unicode: bool = True
    #     ) -> pyspark.rdd.RDD[typing.Tuple[str, str]]:
    # 	pass


__all__ = ["SparkContext"]



# This code is based on code from Apache Spark under the license found in the LICENSE
# file located in the 'spark' folder.

import calendar
import datetime
import math
import re
import time
from builtins import tuple
from collections.abc import Iterator, Mapping
from types import MappingProxyType
from typing import Any, ClassVar, NoReturn, TypeVar, Union, cast, overload

from typing_extensions import Self

import duckdb
from duckdb.sqltypes import DuckDBPyType

from ..exception import ContributionsAcceptedError

T = TypeVar("T")
U = TypeVar("U")

__all__ = [
    "ArrayType",
    "BinaryType",
    "BitstringType",
    "BooleanType",
    "ByteType",
    "DataType",
    "DateType",
    "DayTimeIntervalType",
    "DecimalType",
    "DoubleType",
    "FloatType",
    "HugeIntegerType",
    "IntegerType",
    "LongType",
    "MapType",
    "NullType",
    "Row",
    "ShortType",
    "StringType",
    "StructField",
    "StructType",
    "TimeNSType",
    "TimeNTZType",
    "TimeType",
    "TimestampMillisecondNTZType",
    "TimestampNTZType",
    "TimestampNanosecondNTZType",
    "TimestampSecondNTZType",
    "TimestampType",
    "UUIDType",
    "UnsignedByteType",
    "UnsignedHugeIntegerType",
    "UnsignedIntegerType",
    "UnsignedLongType",
    "UnsignedShortType",
    "VariantType",
]


class DataType:
    """Base class for data types."""

    def __init__(self, duckdb_type: DuckDBPyType) -> None:  # noqa: D107
        self.duckdb_type = duckdb_type

    def __repr__(self) -> str:  # noqa: D105
        return self.__class__.__name__ + "()"

    def __hash__(self) -> int:  # noqa: D105
        return hash(str(self))

    def __eq__(self, other: object) -> bool:  # noqa: D105
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other: object) -> bool:  # noqa: D105
        return not self.__eq__(other)

    @classmethod
    def typeName(cls) -> str:  # noqa: D102
        return cls.__name__[:-4].lower()

    def simpleString(self) -> str:  # noqa: D102
        return self.typeName()

    def jsonValue(self) -> str | dict[str, Any]:  # noqa: D102
        raise ContributionsAcceptedError

    def json(self) -> str:  # noqa: D102
        raise ContributionsAcceptedError

    def needConversion(self) -> bool:
        """Does this type needs conversion between Python object and internal SQL object.

        This is used to avoid the unnecessary conversion for ArrayType/MapType/StructType.
        """
        return False

    def toInternal(self, obj: Any) -> Any:  # noqa: ANN401
        """Converts a Python object into an internal SQL object."""
        return obj

    def fromInternal(self, obj: Any) -> Any:  # noqa: ANN401
        """Converts an internal SQL object into a native Python object."""
        return obj


# This singleton pattern does not work with pickle, you will get
# another object after pickle and unpickle
class DataTypeSingleton(type):
    """Metaclass for DataType."""

    _instances: ClassVar[dict[type["DataTypeSingleton"], "DataTypeSingleton"]] = {}

    def __call__(cls: type[T]) -> T:  # type: ignore[override]
        if cls not in cls._instances:  # type: ignore[attr-defined]
            cls._instances[cls] = super().__call__()  # type: ignore[misc, attr-defined]
        return cls._instances[cls]  # type: ignore[attr-defined]


class NullType(DataType, metaclass=DataTypeSingleton):
    """Null type.

    The data type representing None, used for the types that cannot be inferred.
    """

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("NULL"))

    @classmethod
    def typeName(cls) -> str:  # noqa: D102
        return "void"


class AtomicType(DataType):
    """An internal type used to represent everything that is not
    null, UDTs, arrays, structs, and maps.
    """  # noqa: D205


class NumericType(AtomicType):
    """Numeric data types."""


class IntegralType(NumericType, metaclass=DataTypeSingleton):
    """Integral data types."""


class FractionalType(NumericType):
    """Fractional data types."""


class StringType(AtomicType, metaclass=DataTypeSingleton):
    """String data type."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("VARCHAR"))


class BitstringType(AtomicType, metaclass=DataTypeSingleton):
    """Bitstring data type."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("BIT"))


class UUIDType(AtomicType, metaclass=DataTypeSingleton):
    """UUID data type."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("UUID"))


class BinaryType(AtomicType, metaclass=DataTypeSingleton):
    """Binary (byte array) data type."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("BLOB"))


class BooleanType(AtomicType, metaclass=DataTypeSingleton):
    """Boolean data type."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("BOOLEAN"))


class VariantType(AtomicType, metaclass=DataTypeSingleton):
    """Variant (semi-structured) data type."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("VARIANT"))


class DateType(AtomicType, metaclass=DataTypeSingleton):
    """Date (datetime.date) data type."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("DATE"))

    EPOCH_ORDINAL = datetime.datetime(1970, 1, 1).toordinal()

    def needConversion(self) -> bool:  # noqa: D102
        return True

    def toInternal(self, d: datetime.date) -> int:  # noqa: D102
        if d is not None:
            return d.toordinal() - self.EPOCH_ORDINAL

    def fromInternal(self, v: int) -> datetime.date:  # noqa: D102
        if v is not None:
            return datetime.date.fromordinal(v + self.EPOCH_ORDINAL)


class TimestampType(AtomicType, metaclass=DataTypeSingleton):
    """Timestamp (datetime.datetime) data type."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("TIMESTAMPTZ"))

    @classmethod
    def typeName(cls) -> str:  # noqa: D102
        return "timestamptz"

    def needConversion(self) -> bool:  # noqa: D102
        return True

    def toInternal(self, dt: datetime.datetime) -> int:  # noqa: D102
        if dt is not None:
            seconds = calendar.timegm(dt.utctimetuple()) if dt.tzinfo else time.mktime(dt.timetuple())
            return int(seconds) * 1000000 + dt.microsecond

    def fromInternal(self, ts: int) -> datetime.datetime:  # noqa: D102
        if ts is not None:
            # using int to avoid precision loss in float
            return datetime.datetime.fromtimestamp(ts // 1000000).replace(microsecond=ts % 1000000)


class TimestampNTZType(AtomicType, metaclass=DataTypeSingleton):
    """Timestamp (datetime.datetime) data type without timezone information with microsecond precision."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("TIMESTAMP"))

    def needConversion(self) -> bool:  # noqa: D102
        return True

    @classmethod
    def typeName(cls) -> str:  # noqa: D102
        return "timestamp"

    def toInternal(self, dt: datetime.datetime) -> int:  # noqa: D102
        if dt is not None:
            seconds = calendar.timegm(dt.timetuple())
            return int(seconds) * 1000000 + dt.microsecond

    def fromInternal(self, ts: int) -> datetime.datetime:  # noqa: D102
        if ts is not None:
            # using int to avoid precision loss in float
            return datetime.datetime.utcfromtimestamp(ts // 1000000).replace(microsecond=ts % 1000000)


class TimestampSecondNTZType(AtomicType, metaclass=DataTypeSingleton):
    """Timestamp (datetime.datetime) data type without timezone information with second precision."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("TIMESTAMP_S"))

    def needConversion(self) -> bool:  # noqa: D102
        return True

    @classmethod
    def typeName(cls) -> str:  # noqa: D102
        return "timestamp_s"

    def toInternal(self, dt: datetime.datetime) -> int:  # noqa: D102
        raise ContributionsAcceptedError

    def fromInternal(self, ts: int) -> datetime.datetime:  # noqa: D102
        raise ContributionsAcceptedError


class TimestampMillisecondNTZType(AtomicType, metaclass=DataTypeSingleton):
    """Timestamp (datetime.datetime) data type without timezone information with millisecond precision."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("TIMESTAMP_MS"))

    def needConversion(self) -> bool:  # noqa: D102
        return True

    @classmethod
    def typeName(cls) -> str:  # noqa: D102
        return "timestamp_ms"

    def toInternal(self, dt: datetime.datetime) -> int:  # noqa: D102
        raise ContributionsAcceptedError

    def fromInternal(self, ts: int) -> datetime.datetime:  # noqa: D102
        raise ContributionsAcceptedError


class TimestampNanosecondNTZType(AtomicType, metaclass=DataTypeSingleton):
    """Timestamp (datetime.datetime) data type without timezone information with nanosecond precision."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("TIMESTAMP_NS"))

    def needConversion(self) -> bool:  # noqa: D102
        return True

    @classmethod
    def typeName(cls) -> str:  # noqa: D102
        return "timestamp_ns"

    def toInternal(self, dt: datetime.datetime) -> int:  # noqa: D102
        raise ContributionsAcceptedError

    def fromInternal(self, ts: int) -> datetime.datetime:  # noqa: D102
        raise ContributionsAcceptedError


class DecimalType(FractionalType):
    """Decimal (decimal.Decimal) data type.

    The DecimalType must have fixed precision (the maximum total number of digits)
    and scale (the number of digits on the right of dot). For example, (5, 2) can
    support the value from [-999.99 to 999.99].

    The precision can be up to 38, the scale must be less or equal to precision.

    When creating a DecimalType, the default precision and scale is (10, 0). When inferring
    schema from decimal.Decimal objects, it will be DecimalType(38, 18).

    Parameters
    ----------
    precision : int, optional
        the maximum (i.e. total) number of digits (default: 10)
    scale : int, optional
        the number of digits on right side of dot. (default: 0)
    """

    def __init__(self, precision: int = 10, scale: int = 0) -> None:  # noqa: D107
        super().__init__(duckdb.decimal_type(precision, scale))
        self.precision = precision
        self.scale = scale
        self.hasPrecisionInfo = True  # this is a public API

    def simpleString(self) -> str:  # noqa: D102
        return f"decimal({int(self.precision):d},{int(self.scale):d})"

    def __repr__(self) -> str:  # noqa: D105
        return f"DecimalType({int(self.precision):d},{int(self.scale):d})"


class DoubleType(FractionalType, metaclass=DataTypeSingleton):
    """Double data type, representing double precision floats."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("DOUBLE"))


class FloatType(FractionalType, metaclass=DataTypeSingleton):
    """Float data type, representing single precision floats."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("FLOAT"))


class ByteType(IntegralType):
    """Byte data type, i.e. a signed integer in a single byte."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("TINYINT"))

    def simpleString(self) -> str:  # noqa: D102
        return "tinyint"


class UnsignedByteType(IntegralType):
    """Unsigned byte data type, i.e. a unsigned integer in a single byte."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("UTINYINT"))

    def simpleString(self) -> str:  # noqa: D102
        return "utinyint"


class ShortType(IntegralType):
    """Short data type, i.e. a signed 16-bit integer."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("SMALLINT"))

    def simpleString(self) -> str:  # noqa: D102
        return "smallint"


class UnsignedShortType(IntegralType):
    """Unsigned short data type, i.e. a unsigned 16-bit integer."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("USMALLINT"))

    def simpleString(self) -> str:  # noqa: D102
        return "usmallint"


class IntegerType(IntegralType):
    """Int data type, i.e. a signed 32-bit integer."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("INTEGER"))

    def simpleString(self) -> str:  # noqa: D102
        return "integer"


class UnsignedIntegerType(IntegralType):
    """Unsigned int data type, i.e. a unsigned 32-bit integer."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("UINTEGER"))

    def simpleString(self) -> str:  # noqa: D102
        return "uinteger"


class LongType(IntegralType):
    """Long data type, i.e. a signed 64-bit integer.

    If the values are beyond the range of [-9223372036854775808, 9223372036854775807],
    please use :class:`DecimalType`.
    """

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("BIGINT"))

    def simpleString(self) -> str:  # noqa: D102
        return "bigint"


class UnsignedLongType(IntegralType):
    """Unsigned long data type, i.e. a unsigned 64-bit integer.

    If the values are beyond the range of [0, 18446744073709551615],
    please use :class:`HugeIntegerType`.
    """

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("UBIGINT"))

    def simpleString(self) -> str:  # noqa: D102
        return "ubigint"


class HugeIntegerType(IntegralType):
    """Huge integer data type, i.e. a signed 128-bit integer.

    If the values are beyond the range of [-170141183460469231731687303715884105728,
    170141183460469231731687303715884105727], please use :class:`DecimalType`.
    """

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("HUGEINT"))

    def simpleString(self) -> str:  # noqa: D102
        return "hugeint"


class UnsignedHugeIntegerType(IntegralType):
    """Unsigned huge integer data type, i.e. a unsigned 128-bit integer.

    If the values are beyond the range of [0, 340282366920938463463374607431768211455],
    please use :class:`DecimalType`.
    """

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("UHUGEINT"))

    def simpleString(self) -> str:  # noqa: D102
        return "uhugeint"


class TimeType(IntegralType):
    """Time (datetime.time) data type."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("TIMETZ"))

    def simpleString(self) -> str:  # noqa: D102
        return "timetz"


class TimeNTZType(IntegralType):
    """Time (datetime.time) data type without timezone information."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("TIME"))

    def simpleString(self) -> str:  # noqa: D102
        return "time"


class TimeNSType(IntegralType):
    """Time NS (datetime.time) data type without timezone information."""

    def __init__(self) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("TIME_NS"))

    def simpleString(self) -> str:  # noqa: D102
        return "time_ns"


class DayTimeIntervalType(AtomicType):
    """DayTimeIntervalType (datetime.timedelta)."""

    DAY = 0
    HOUR = 1
    MINUTE = 2
    SECOND = 3

    _fields: Mapping[str, int] = MappingProxyType(
        {
            DAY: "day",
            HOUR: "hour",
            MINUTE: "minute",
            SECOND: "second",
        }
    )

    _inverted_fields: Mapping[int, str] = MappingProxyType(dict(zip(_fields.values(), _fields.keys(), strict=False)))

    def __init__(self, startField: int | None = None, endField: int | None = None) -> None:  # noqa: D107
        super().__init__(DuckDBPyType("INTERVAL"))
        if startField is None and endField is None:
            # Default matched to scala side.
            startField = DayTimeIntervalType.DAY
            endField = DayTimeIntervalType.SECOND
        elif startField is not None and endField is None:
            endField = startField

        fields = DayTimeIntervalType._fields
        if startField not in fields or endField not in fields:
            msg = f"interval {startField} to {endField} is invalid"
            raise RuntimeError(msg)
        self.startField = cast("int", startField)
        self.endField = cast("int", endField)

    def _str_repr(self) -> str:
        fields = DayTimeIntervalType._fields
        start_field_name = fields[self.startField]
        end_field_name = fields[self.endField]
        if start_field_name == end_field_name:
            return f"interval {start_field_name}"
        else:
            return f"interval {start_field_name} to {end_field_name}"

    simpleString = _str_repr

    def __repr__(self) -> str:  # noqa: D105
        return f"{type(self).__name__}({int(self.startField):d}, {int(self.endField):d})"

    def needConversion(self) -> bool:  # noqa: D102
        return True

    def toInternal(self, dt: datetime.timedelta) -> int | None:  # noqa: D102
        if dt is not None:
            return (math.floor(dt.total_seconds()) * 1000000) + dt.microseconds

    def fromInternal(self, micros: int) -> datetime.timedelta | None:  # noqa: D102
        if micros is not None:
            return datetime.timedelta(microseconds=micros)


class ArrayType(DataType):
    """Array data type.

    Parameters
    ----------
    elementType : :class:`DataType`
        :class:`DataType` of each element in the array.
    containsNull : bool, optional
        whether the array can contain null (None) values.

    Examples:
    --------
    >>> ArrayType(StringType()) == ArrayType(StringType(), True)
    True
    >>> ArrayType(StringType(), False) == ArrayType(StringType())
    False
    """

    def __init__(self, elementType: DataType, containsNull: bool = True) -> None:  # noqa: D107
        super().__init__(duckdb.list_type(elementType.duckdb_type))
        assert isinstance(elementType, DataType), f"elementType {elementType} should be an instance of {DataType}"
        self.elementType = elementType
        self.containsNull = containsNull

    def simpleString(self) -> str:  # noqa: D102
        return f"array<{self.elementType.simpleString()}>"

    def __repr__(self) -> str:  # noqa: D105
        return f"ArrayType({self.elementType}, {self.containsNull!s})"

    def needConversion(self) -> bool:  # noqa: D102
        return self.elementType.needConversion()

    def toInternal(self, obj: list[T | None]) -> list[T | None]:  # noqa: D102
        if not self.needConversion():
            return obj
        return obj and [self.elementType.toInternal(v) for v in obj]

    def fromInternal(self, obj: list[T | None]) -> list[T | None]:  # noqa: D102
        if not self.needConversion():
            return obj
        return obj and [self.elementType.fromInternal(v) for v in obj]


class MapType(DataType):
    """Map data type.

    Parameters
    ----------
    keyType : :class:`DataType`
        :class:`DataType` of the keys in the map.
    valueType : :class:`DataType`
        :class:`DataType` of the values in the map.
    valueContainsNull : bool, optional
        indicates whether values can contain null (None) values.

    Notes:
    -----
    Keys in a map data type are not allowed to be null (None).

    Examples:
    --------
    >>> (MapType(StringType(), IntegerType()) == MapType(StringType(), IntegerType(), True))
    True
    >>> (MapType(StringType(), IntegerType(), False) == MapType(StringType(), FloatType()))
    False
    """

    def __init__(self, keyType: DataType, valueType: DataType, valueContainsNull: bool = True) -> None:  # noqa: D107
        super().__init__(duckdb.map_type(keyType.duckdb_type, valueType.duckdb_type))
        assert isinstance(keyType, DataType), f"keyType {keyType} should be an instance of {DataType}"
        assert isinstance(valueType, DataType), f"valueType {valueType} should be an instance of {DataType}"
        self.keyType = keyType
        self.valueType = valueType
        self.valueContainsNull = valueContainsNull

    def simpleString(self) -> str:  # noqa: D102
        return f"map<{self.keyType.simpleString()},{self.valueType.simpleString()}>"

    def __repr__(self) -> str:  # noqa: D105
        return f"MapType({self.keyType}, {self.valueType}, {self.valueContainsNull!s})"

    def needConversion(self) -> bool:  # noqa: D102
        return self.keyType.needConversion() or self.valueType.needConversion()

    def toInternal(self, obj: dict[T, U | None]) -> dict[T, U | None]:  # noqa: D102
        if not self.needConversion():
            return obj
        return obj and {self.keyType.toInternal(k): self.valueType.toInternal(v) for k, v in obj.items()}

    def fromInternal(self, obj: dict[T, U | None]) -> dict[T, U | None]:  # noqa: D102
        if not self.needConversion():
            return obj
        return obj and {self.keyType.fromInternal(k): self.valueType.fromInternal(v) for k, v in obj.items()}


class StructField(DataType):
    """A field in :class:`StructType`.

    Parameters
    ----------
    name : str
        name of the field.
    dataType : :class:`DataType`
        :class:`DataType` of the field.
    nullable : bool, optional
        whether the field can be null (None) or not.
    metadata : dict, optional
        a dict from string to simple type that can be toInternald to JSON automatically

    Examples:
    --------
    >>> (StructField("f1", StringType(), True) == StructField("f1", StringType(), True))
    True
    >>> (StructField("f1", StringType(), True) == StructField("f2", StringType(), True))
    False
    """

    def __init__(  # noqa: D107
        self,
        name: str,
        dataType: DataType,
        nullable: bool = True,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        super().__init__(dataType.duckdb_type)
        assert isinstance(dataType, DataType), f"dataType {dataType} should be an instance of {DataType}"
        assert isinstance(name, str), f"field name {name} should be a string"
        self.name = name
        self.dataType = dataType
        self.nullable = nullable
        self.metadata = metadata or {}

    def simpleString(self) -> str:  # noqa: D102
        return f"{self.name}:{self.dataType.simpleString()}"

    def __repr__(self) -> str:  # noqa: D105
        return f"StructField('{self.name}', {self.dataType}, {self.nullable!s})"

    def needConversion(self) -> bool:  # noqa: D102
        return self.dataType.needConversion()

    def toInternal(self, obj: T) -> T:  # noqa: D102
        return self.dataType.toInternal(obj)

    def fromInternal(self, obj: T) -> T:  # noqa: D102
        return self.dataType.fromInternal(obj)

    def typeName(self) -> str:  # type: ignore[override]  # noqa: D102
        msg = "StructField does not have typeName. Use typeName on its type explicitly instead."
        raise TypeError(msg)


class StructType(DataType):
    r"""Struct type, consisting of a list of :class:`StructField`.

    This is the data type representing a :class:`Row`.

    Iterating a :class:`StructType` will iterate over its :class:`StructField`\\s.
    A contained :class:`StructField` can be accessed by its name or position.

    Examples:
    --------
    >>> struct1 = StructType([StructField("f1", StringType(), True)])
    >>> struct1["f1"]
    StructField('f1', StringType(), True)
    >>> struct1[0]
    StructField('f1', StringType(), True)

    >>> struct1 = StructType([StructField("f1", StringType(), True)])
    >>> struct2 = StructType([StructField("f1", StringType(), True)])
    >>> struct1 == struct2
    True
    >>> struct1 = StructType([StructField("f1", StringType(), True)])
    >>> struct2 = StructType(
    ...     [StructField("f1", StringType(), True), StructField("f2", IntegerType(), False)]
    ... )
    >>> struct1 == struct2
    False
    """

    def _update_internal_duckdb_type(self) -> None:
        self.duckdb_type = duckdb.struct_type(dict(zip(self.names, [x.duckdb_type for x in self.fields], strict=False)))

    def __init__(self, fields: list[StructField] | None = None) -> None:  # noqa: D107
        if not fields:
            self.fields = []
            self.names = []
        else:
            self.fields = fields
            self.names = [f.name for f in fields]
            assert all(isinstance(f, StructField) for f in fields), "fields should be a list of StructField"
        # Precalculated list of fields that need conversion with fromInternal/toInternal functions
        self._needConversion = [f.needConversion() for f in self]
        self._needSerializeAnyField = any(self._needConversion)
        super().__init__(duckdb.struct_type(dict(zip(self.names, [x.duckdb_type for x in self.fields], strict=False))))

    @overload
    def add(
        self,
        field: str,
        data_type: str | DataType,
        nullable: bool = True,
        metadata: dict[str, Any] | None = None,
    ) -> "StructType": ...

    @overload
    def add(self, field: StructField) -> "StructType": ...

    def add(
        self,
        field: str | StructField,
        data_type: str | DataType | None = None,
        nullable: bool = True,
        metadata: dict[str, Any] | None = None,
    ) -> "StructType":
        r"""Construct a :class:`StructType` by adding new elements to it, to define the schema.
        The method accepts either:

            a) A single parameter which is a :class:`StructField` object.
            b) Between 2 and 4 parameters as (name, data_type, nullable (optional),
               metadata(optional). The data_type parameter may be either a String or a
               :class:`DataType` object.

        Parameters
        ----------
        field : str or :class:`StructField`
            Either the name of the field or a :class:`StructField` object
        data_type : :class:`DataType`, optional
            If present, the DataType of the :class:`StructField` to create
        nullable : bool, optional
            Whether the field to add should be nullable (default True)
        metadata : dict, optional
            Any additional metadata (default None)

        Returns:
        -------
        :class:`StructType`

        Examples:
        --------
        >>> struct1 = StructType().add("f1", StringType(), True).add("f2", StringType(), True, None)
        >>> struct2 = StructType([StructField("f1", StringType(), True), \\
        ...     StructField("f2", StringType(), True, None)])
        >>> struct1 == struct2
        True
        >>> struct1 = StructType().add(StructField("f1", StringType(), True))
        >>> struct2 = StructType([StructField("f1", StringType(), True)])
        >>> struct1 == struct2
        True
        >>> struct1 = StructType().add("f1", "string", True)
        >>> struct2 = StructType([StructField("f1", StringType(), True)])
        >>> struct1 == struct2
        True
        """  # noqa: D205, D415
        if isinstance(field, StructField):
            self.fields.append(field)
            self.names.append(field.name)
        else:
            if isinstance(field, str) and data_type is None:
                msg = "Must specify DataType if passing name of struct_field to create."
                raise ValueError(msg)
            else:
                data_type_f = data_type
            self.fields.append(StructField(field, data_type_f, nullable, metadata))
            self.names.append(field)
        # Precalculated list of fields that need conversion with fromInternal/toInternal functions
        self._needConversion = [f.needConversion() for f in self]
        self._needSerializeAnyField = any(self._needConversion)
        self._update_internal_duckdb_type()
        return self

    def __iter__(self) -> Iterator[StructField]:
        """Iterate the fields."""
        return iter(self.fields)

    def __len__(self) -> int:
        """Return the number of fields."""
        return len(self.fields)

    def __getitem__(self, key: str | int) -> StructField:
        """Access fields by name or slice."""
        if isinstance(key, str):
            for field in self:
                if field.name == key:
                    return field
            msg = f"No StructField named {key}"
            raise KeyError(msg)
        elif isinstance(key, int):
            try:
                return self.fields[key]
            except IndexError:
                msg = "StructType index out of range"
                raise IndexError(msg)  # noqa: B904
        elif isinstance(key, slice):
            return StructType(self.fields[key])
        else:
            msg = "StructType keys should be strings, integers or slices"
            raise TypeError(msg)

    def simpleString(self) -> str:  # noqa: D102
        return "struct<{}>".format(",".join(f.simpleString() for f in self))

    def __repr__(self) -> str:  # noqa: D105
        return "StructType([{}])".format(", ".join(str(field) for field in self))

    def __contains__(self, item: str) -> bool:  # noqa: D105
        return item in self.names

    def extract_types_and_names(self) -> tuple[list[str], list[str]]:  # noqa: D102
        names = []
        types = []
        for f in self.fields:
            types.append(str(f.dataType.duckdb_type))
            names.append(f.name)
        return (types, names)

    def fieldNames(self) -> list[str]:
        """Returns all field names in a list.

        Examples:
        --------
        >>> struct = StructType([StructField("f1", StringType(), True)])
        >>> struct.fieldNames()
        ['f1']
        """
        return list(self.names)

    def treeString(self, level: int | None = None) -> str:
        """Returns a string representation of the schema in tree format.

        Parameters
        ----------
        level : int, optional
            Maximum depth to print. If None, prints all levels.

        Returns:
        -------
        str
            Tree-formatted schema string

        Examples:
        --------
        >>> schema = StructType([StructField("age", IntegerType(), True)])
        >>> print(schema.treeString())
        root
         |-- age: integer (nullable = true)
        """

        def _tree_string(schema: "StructType", depth: int = 0, max_depth: int | None = None) -> list[str]:
            """Recursively build tree string lines."""
            lines = []
            if depth == 0:
                lines.append("root")

            if max_depth is not None and depth >= max_depth:
                return lines

            for field in schema.fields:
                indent = " " * depth
                prefix = " |-- "
                nullable_str = "true" if field.nullable else "false"

                # Handle nested StructType
                if isinstance(field.dataType, StructType):
                    lines.append(f"{indent}{prefix}{field.name}: struct (nullable = {nullable_str})")
                    # Recursively handle nested struct - don't skip any lines, root only appears at depth 0
                    nested_lines = _tree_string(field.dataType, depth + 1, max_depth)
                    lines.extend(nested_lines)
                # Handle ArrayType
                elif isinstance(field.dataType, ArrayType):
                    element_type = field.dataType.elementType
                    if isinstance(element_type, StructType):
                        lines.append(f"{indent}{prefix}{field.name}: array (nullable = {nullable_str})")
                        lines.append(
                            f"{indent} |    |-- element: struct (containsNull = {field.dataType.containsNull})"
                        )
                        nested_lines = _tree_string(element_type, depth + 2, max_depth)
                        lines.extend(nested_lines)
                    else:
                        type_str = element_type.simpleString()
                        lines.append(f"{indent}{prefix}{field.name}: array<{type_str}> (nullable = {nullable_str})")
                # Handle MapType
                elif isinstance(field.dataType, MapType):
                    key_type = field.dataType.keyType.simpleString()
                    value_type = field.dataType.valueType.simpleString()
                    lines.append(
                        f"{indent}{prefix}{field.name}: map<{key_type},{value_type}> (nullable = {nullable_str})"
                    )
                # Handle simple types
                else:
                    type_str = field.dataType.simpleString()
                    lines.append(f"{indent}{prefix}{field.name}: {type_str} (nullable = {nullable_str})")

            return lines

        lines = _tree_string(self, 0, level)
        return "\n".join(lines)

    def needConversion(self) -> bool:  # noqa: D102
        # We need convert Row()/namedtuple into tuple()
        return True

    def toInternal(self, obj: tuple) -> tuple:  # noqa: D102
        if obj is None:
            return

        if self._needSerializeAnyField:
            # Only calling toInternal function for fields that need conversion
            if isinstance(obj, dict):
                return tuple(
                    f.toInternal(obj.get(n)) if c else obj.get(n)
                    for n, f, c in zip(self.names, self.fields, self._needConversion, strict=False)
                )
            elif isinstance(obj, (tuple, list)):
                return tuple(
                    f.toInternal(v) if c else v for f, v, c in zip(self.fields, obj, self._needConversion, strict=False)
                )
            elif hasattr(obj, "__dict__"):
                d = obj.__dict__
                return tuple(
                    f.toInternal(d.get(n)) if c else d.get(n)
                    for n, f, c in zip(self.names, self.fields, self._needConversion, strict=False)
                )
            else:
                msg = f"Unexpected tuple {obj!r} with StructType"
                raise ValueError(msg)
        else:
            if isinstance(obj, dict):
                return tuple(obj.get(n) for n in self.names)
            elif isinstance(obj, (list, tuple)):
                return tuple(obj)
            elif hasattr(obj, "__dict__"):
                d = obj.__dict__
                return tuple(d.get(n) for n in self.names)
            else:
                msg = f"Unexpected tuple {obj!r} with StructType"
                raise ValueError(msg)

    def fromInternal(self, obj: tuple) -> "Row":  # noqa: D102
        if obj is None:
            return
        if isinstance(obj, Row):
            # it's already converted by pickler
            return obj

        values: tuple | list
        if self._needSerializeAnyField:
            # Only calling fromInternal function for fields that need conversion
            values = [
                f.fromInternal(v) if c else v for f, v, c in zip(self.fields, obj, self._needConversion, strict=False)
            ]
        else:
            values = obj
        return _create_row(self.names, values)


class UnionType(DataType):
    def __init__(self) -> None:
        raise ContributionsAcceptedError


class UserDefinedType(DataType):
    """User-defined type (UDT).

    .. note:: WARN: Spark Internal Use Only
    """

    def __init__(self) -> None:
        raise ContributionsAcceptedError

    @classmethod
    def typeName(cls) -> str:
        return cls.__name__.lower()

    @classmethod
    def sqlType(cls) -> DataType:
        """Underlying SQL storage type for this UDT."""
        msg = "UDT must implement sqlType()."
        raise NotImplementedError(msg)

    @classmethod
    def module(cls) -> str:
        """The Python module of the UDT."""
        msg = "UDT must implement module()."
        raise NotImplementedError(msg)

    @classmethod
    def scalaUDT(cls) -> str:
        """The class name of the paired Scala UDT (could be '', if there
        is no corresponding one).
        """  # noqa: D205
        return ""

    def needConversion(self) -> bool:
        return True

    @classmethod
    def _cachedSqlType(cls) -> DataType:
        """Cache the sqlType() into class, because it's heavily used in `toInternal`."""
        if not hasattr(cls, "_cached_sql_type"):
            cls._cached_sql_type = cls.sqlType()  # type: ignore[attr-defined]
        return cls._cached_sql_type  # type: ignore[attr-defined]

    def toInternal(self, obj: Any) -> Any:  # noqa: ANN401
        if obj is not None:
            return self._cachedSqlType().toInternal(self.serialize(obj))

    def fromInternal(self, obj: Any) -> Any:  # noqa: ANN401
        v = self._cachedSqlType().fromInternal(obj)
        if v is not None:
            return self.deserialize(v)

    def serialize(self, obj: Any) -> NoReturn:  # noqa: ANN401
        """Converts a user-type object into a SQL datum."""
        msg = "UDT must implement toInternal()."
        raise NotImplementedError(msg)

    def deserialize(self, datum: Any) -> NoReturn:  # noqa: ANN401
        """Converts a SQL datum into a user-type object."""
        msg = "UDT must implement fromInternal()."
        raise NotImplementedError(msg)

    def simpleString(self) -> str:
        return "udt"

    def __eq__(self, other: object) -> bool:
        return type(self) is type(other)


_atomic_types: list[type[DataType]] = [
    StringType,
    BinaryType,
    BooleanType,
    DecimalType,
    FloatType,
    DoubleType,
    ByteType,
    ShortType,
    IntegerType,
    LongType,
    DateType,
    TimestampType,
    TimestampNTZType,
    NullType,
]
_all_atomic_types: dict[str, type[DataType]] = {t.typeName(): t for t in _atomic_types}

_complex_types: list[type[ArrayType | MapType | StructType]] = [
    ArrayType,
    MapType,
    StructType,
]
_all_complex_types: dict[str, type[ArrayType | MapType | StructType]] = {v.typeName(): v for v in _complex_types}


_FIXED_DECIMAL = re.compile(r"decimal\(\s*(\d+)\s*,\s*(-?\d+)\s*\)")
_INTERVAL_DAYTIME = re.compile(r"interval (day|hour|minute|second)( to (day|hour|minute|second))?")


def _create_row(fields: Union["Row", list[str]], values: tuple[Any, ...] | list[Any]) -> "Row":
    row = Row(*values)
    row.__fields__ = fields
    return row


class Row(tuple):
    """A row in :class:`DataFrame`.
    The fields in it can be accessed:

    * like attributes (``row.key``)
    * like dictionary values (``row[key]``)

    ``key in row`` will search through row keys.

    Row can be used to create a row object by using named arguments.
    It is not allowed to omit a named argument to represent that the value is
    None or missing. This should be explicitly set to None in this case.

    .. versionchanged:: 3.0.0
        Rows created from named arguments no longer have
        field names sorted alphabetically and will be ordered in the position as
        entered.

    Examples:
    --------
    >>> row = Row(name="Alice", age=11)
    >>> row
    Row(name='Alice', age=11)
    >>> row["name"], row["age"]
    ('Alice', 11)
    >>> row.name, row.age
    ('Alice', 11)
    >>> "name" in row
    True
    >>> "wrong_key" in row
    False

    Row also can be used to create another Row like class, then it
    could be used to create Row objects, such as

    >>> Person = Row("name", "age")
    >>> Person
    <Row('name', 'age')>
    >>> "name" in Person
    True
    >>> "wrong_key" in Person
    False
    >>> Person("Alice", 11)
    Row(name='Alice', age=11)

    This form can also be used to create rows as tuple values, i.e. with unnamed
    fields.

    >>> row1 = Row("Alice", 11)
    >>> row2 = Row(name="Alice", age=11)
    >>> row1 == row2
    True
    """  # noqa: D205, D415

    @overload
    def __new__(cls, *args: str) -> "Row": ...

    @overload
    def __new__(cls, **kwargs: Any) -> "Row": ...  # noqa: ANN401

    def __new__(cls, *args: str | None, **kwargs: Any | None) -> Self:  # noqa: D102
        if args and kwargs:
            msg = "Can not use both args and kwargs to create Row"
            raise ValueError(msg)
        if kwargs:
            # create row objects
            row = tuple.__new__(cls, list(kwargs.values()))
            row.__fields__ = list(kwargs.keys())
            return row
        # create row class or objects
        return tuple.__new__(cls, args)

    def asDict(self, recursive: bool = False) -> dict[str, Any]:
        """Return as a dict.

        Parameters
        ----------
        recursive : bool, optional
            turns the nested Rows to dict (default: False).

        Notes:
        -----
        If a row contains duplicate field names, e.g., the rows of a join
        between two :class:`DataFrame` that both have the fields of same names,
        one of the duplicate fields will be selected by ``asDict``. ``__getitem__``
        will also return one of the duplicate fields, however returned value might
        be different to ``asDict``.

        Examples:
        --------
        >>> Row(name="Alice", age=11).asDict() == {"name": "Alice", "age": 11}
        True
        >>> row = Row(key=1, value=Row(name="a", age=2))
        >>> row.asDict() == {"key": 1, "value": Row(name="a", age=2)}
        True
        >>> row.asDict(True) == {"key": 1, "value": {"name": "a", "age": 2}}
        True
        """
        if not hasattr(self, "__fields__"):
            msg = "Cannot convert a Row class into dict"
            raise TypeError(msg)

        if recursive:

            def conv(obj: Row | list | dict | object) -> list | dict | object:
                if isinstance(obj, Row):
                    return obj.asDict(True)
                elif isinstance(obj, list):
                    return [conv(o) for o in obj]
                elif isinstance(obj, dict):
                    return {k: conv(v) for k, v in obj.items()}
                else:
                    return obj

            return dict(zip(self.__fields__, (conv(o) for o in self), strict=False))
        else:
            return dict(zip(self.__fields__, self, strict=False))

    def __contains__(self, item: Any) -> bool:  # noqa: D105, ANN401
        if hasattr(self, "__fields__"):
            return item in self.__fields__
        else:
            return super().__contains__(item)

    # let object acts like class
    def __call__(self, *args: Any) -> "Row":  # noqa: ANN401
        """Create new Row object."""
        if len(args) > len(self):
            msg = f"Can not create Row with fields {self}, expected {len(self):d} values but got {args}"
            raise ValueError(msg)
        return _create_row(self, args)

    def __getitem__(self, item: Any) -> Any:  # noqa: D105, ANN401
        if isinstance(item, (int, slice)):
            return super().__getitem__(item)
        try:
            # it will be slow when it has many fields,
            # but this will not be used in normal cases
            idx = self.__fields__.index(item)
            return super().__getitem__(idx)
        except IndexError:
            raise KeyError(item)  # noqa: B904
        except ValueError:
            raise ValueError(item)  # noqa: B904

    def __getattr__(self, item: str) -> Any:  # noqa: D105, ANN401
        if item.startswith("__"):
            raise AttributeError(item)
        try:
            # it will be slow when it has many fields,
            # but this will not be used in normal cases
            idx = self.__fields__.index(item)
            return self[idx]
        except IndexError:
            raise AttributeError(item)  # noqa: B904
        except ValueError:
            raise AttributeError(item)  # noqa: B904

    def __setattr__(self, key: Any, value: Any) -> None:  # noqa: D105, ANN401
        if key != "__fields__":
            msg = "Row is read-only"
            raise RuntimeError(msg)
        self.__dict__[key] = value

    def __reduce__(
        self,
    ) -> str | tuple[Any, ...]:
        """Returns a tuple so Python knows how to pickle Row."""
        if hasattr(self, "__fields__"):
            return (_create_row, (self.__fields__, tuple(self)))
        else:
            return tuple.__reduce__(self)

    def __repr__(self) -> str:
        """Printable representation of Row used in Python REPL."""
        if hasattr(self, "__fields__"):
            return "Row({})".format(", ".join(f"{k}={v!r}" for k, v in zip(self.__fields__, tuple(self), strict=False)))
        else:
            return "<Row({})>".format(", ".join(f"{field!r}" for field in self))



"""DuckDB DB API 2.0 Type Objects Module.

This module provides DB API 2.0 compliant type objects for DuckDB, allowing applications
to check column types returned by queries against standard database API categories.

Example:
    >>> import duckdb
    >>>
    >>> conn = duckdb.connect()
    >>> cursor = conn.cursor()
    >>> cursor.execute("SELECT 'hello' as text_col, 42 as num_col, CURRENT_DATE as date_col")
    >>>
    >>> # Check column types using DB API type objects
    >>> for i, desc in enumerate(cursor.description):
    >>>     col_name, col_type = desc[0], desc[1]
    >>>     if col_type == duckdb.STRING:
    >>>         print(f"{col_name} is a string type")
    >>>     elif col_type == duckdb.NUMBER:
    >>>         print(f"{col_name} is a numeric type")
    >>>     elif col_type == duckdb.DATETIME:
    >>>         print(f"{col_name} is a date/time type")

See Also:
    - PEP 249: https://peps.python.org/pep-0249/
    - DuckDB Type System: https://duckdb.org/docs/sql/data_types/overview
"""

from duckdb import sqltypes


class DBAPITypeObject:
    """DB API 2.0 type object for categorizing database column types.

    This class implements the type objects defined in PEP 249 (DB API 2.0).
    It allows checking whether a specific DuckDB type belongs to a broader
    category like STRING, NUMBER, DATETIME, etc.

    The type object supports equality comparison with DuckDBPyType instances,
    returning True if the type belongs to this category.

    Args:
        types: A list of DuckDBPyType instances that belong to this type category.

    Example:
        >>> string_types = DBAPITypeObject([sqltypes.VARCHAR, sqltypes.CHAR])
        >>> result = sqltypes.VARCHAR == string_types  # True
        >>> result = sqltypes.INTEGER == string_types  # False

    Note:
        This follows the DB API 2.0 specification where type objects are compared
        using equality operators rather than isinstance() checks.
    """

    def __init__(self, types: list[sqltypes.DuckDBPyType]) -> None:
        """Initialize a DB API type object.

        Args:
            types: List of DuckDB types that belong to this category.
        """
        self.types = types

    def __eq__(self, other: object) -> bool:
        """Check if a DuckDB type belongs to this type category.

        This method implements the DB API 2.0 type checking mechanism.
        It returns True if the other object is a DuckDBPyType that
        is contained in this type category.

        Args:
            other: The object to compare, typically a DuckDBPyType instance.

        Returns:
            True if other is a DuckDBPyType in this category, False otherwise.

        Example:
            >>> NUMBER == sqltypes.INTEGER  # True
            >>> NUMBER == sqltypes.VARCHAR  # False
        """
        if isinstance(other, sqltypes.DuckDBPyType):
            return other in self.types
        return False

    def __repr__(self) -> str:
        """Return a string representation of this type object.

        Returns:
            A string showing the type object and its contained DuckDB types.

        Example:
            >>> repr(STRING)
            '<DBAPITypeObject [VARCHAR]>'
        """
        return f"<DBAPITypeObject [{','.join(str(x) for x in self.types)}]>"


# Define the standard DB API 2.0 type objects for DuckDB

STRING = DBAPITypeObject([sqltypes.VARCHAR])
"""
STRING type object for text-based database columns.

This type object represents all string/text types in DuckDB. Currently includes:
- VARCHAR: Variable-length character strings

Use this to check if a column contains textual data that should be handled
as Python strings.

DB API 2.0 Reference:
    https://peps.python.org/pep-0249/#string

Example:
    >>> cursor.description[0][1] == STRING  # Check if first column is text
"""

NUMBER = DBAPITypeObject(
    [
        sqltypes.TINYINT,
        sqltypes.UTINYINT,
        sqltypes.SMALLINT,
        sqltypes.USMALLINT,
        sqltypes.INTEGER,
        sqltypes.UINTEGER,
        sqltypes.BIGINT,
        sqltypes.UBIGINT,
        sqltypes.HUGEINT,
        sqltypes.UHUGEINT,
        sqltypes.DuckDBPyType("BIGNUM"),
        sqltypes.DuckDBPyType("DECIMAL"),
        sqltypes.FLOAT,
        sqltypes.DOUBLE,
    ]
)
"""
NUMBER type object for numeric database columns.

This type object represents all numeric types in DuckDB, including:

Integer Types:
- TINYINT, UTINYINT: 8-bit signed/unsigned integers
- SMALLINT, USMALLINT: 16-bit signed/unsigned integers
- INTEGER, UINTEGER: 32-bit signed/unsigned integers
- BIGINT, UBIGINT: 64-bit signed/unsigned integers
- HUGEINT, UHUGEINT: 128-bit signed/unsigned integers

Decimal Types:
- BIGNUM: Arbitrary precision integers
- DECIMAL: Fixed-point decimal numbers

Floating Point Types:
- FLOAT: 32-bit floating point
- DOUBLE: 64-bit floating point

Use this to check if a column contains numeric data that should be handled
as Python int, float, or Decimal objects.

DB API 2.0 Reference:
    https://peps.python.org/pep-0249/#number

Example:
    >>> cursor.description[1][1] == NUMBER  # Check if second column is numeric
"""

DATETIME = DBAPITypeObject(
    [
        sqltypes.DATE,
        sqltypes.TIME,
        sqltypes.TIME_TZ,
        sqltypes.TIMESTAMP,
        sqltypes.TIMESTAMP_TZ,
        sqltypes.TIMESTAMP_NS,
        sqltypes.TIMESTAMP_MS,
        sqltypes.TIMESTAMP_S,
    ]
)
"""
DATETIME type object for date and time database columns.

This type object represents all date/time types in DuckDB, including:

Date Types:
- DATE: Calendar dates (year, month, day)

Time Types:
- TIME: Time of day without timezone
- TIME_TZ: Time of day with timezone

Timestamp Types:
- TIMESTAMP: Date and time without timezone (microsecond precision)
- TIMESTAMP_TZ: Date and time with timezone
- TIMESTAMP_NS: Nanosecond precision timestamps
- TIMESTAMP_MS: Millisecond precision timestamps
- TIMESTAMP_S: Second precision timestamps

Use this to check if a column contains temporal data that should be handled
as Python datetime, date, or time objects.

DB API 2.0 Reference:
    https://peps.python.org/pep-0249/#datetime

Example:
    >>> cursor.description[2][1] == DATETIME  # Check if third column is date/time
"""

BINARY = DBAPITypeObject([sqltypes.BLOB])
"""
BINARY type object for binary data database columns.

This type object represents binary data types in DuckDB:
- BLOB: Binary Large Objects for storing arbitrary binary data

Use this to check if a column contains binary data that should be handled
as Python bytes objects.

DB API 2.0 Reference:
    https://peps.python.org/pep-0249/#binary

Example:
    >>> cursor.description[3][1] == BINARY  # Check if fourth column is binary
"""

ROWID = None
"""
ROWID type object for row identifier columns.

DB API 2.0 Reference:
    https://peps.python.org/pep-0249/#rowid

Note:
    This will always be None for DuckDB connections. Applications should not
    rely on ROWID functionality when using DuckDB.
"""



from typing import ClassVar, Protocol, SupportsInt, Literal, TypeAlias

class CppEnum(Protocol):
    """Base Enum-like Protocol class in C++ code.

    Correspond to `py::enum_` in Pybind11.

    Note:
        This is marked as a `Protocol` to specify that an `isinstance` check against this class won't work, as this is a typing-only construct.
    """
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: SupportsInt) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: SupportsInt) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class CSVLineTerminator(CppEnum):
    CARRIAGE_RETURN_LINE_FEED: ClassVar[CSVLineTerminator]  # value = <CSVLineTerminator.CARRIAGE_RETURN_LINE_FEED: 1>
    LINE_FEED: ClassVar[CSVLineTerminator]  # value = <CSVLineTerminator.LINE_FEED: 0>
    __members__: ClassVar[
        dict[str, CSVLineTerminator]
    ]  # value = {'LINE_FEED': <CSVLineTerminator.LINE_FEED: 0>, 'CARRIAGE_RETURN_LINE_FEED': <CSVLineTerminator.CARRIAGE_RETURN_LINE_FEED: 1>}  # noqa: E501

class ExpectedResultType(CppEnum):
    CHANGED_ROWS: ClassVar[ExpectedResultType]  # value = <ExpectedResultType.CHANGED_ROWS: 1>
    NOTHING: ClassVar[ExpectedResultType]  # value = <ExpectedResultType.NOTHING: 2>
    QUERY_RESULT: ClassVar[ExpectedResultType]  # value = <ExpectedResultType.QUERY_RESULT: 0>
    __members__: ClassVar[
        dict[str, ExpectedResultType]
    ]  # value = {'QUERY_RESULT': <ExpectedResultType.QUERY_RESULT: 0>, 'CHANGED_ROWS': <ExpectedResultType.CHANGED_ROWS: 1>, 'NOTHING': <ExpectedResultType.NOTHING: 2>}  # noqa: E501

class ExplainType(CppEnum):
    ANALYZE: ClassVar[ExplainType]  # value = <ExplainType.ANALYZE: 1>
    STANDARD: ClassVar[ExplainType]  # value = <ExplainType.STANDARD: 0>
    __members__: ClassVar[
        dict[str, ExplainType]
    ]  # value = {'STANDARD': <ExplainType.STANDARD: 0>, 'ANALYZE': <ExplainType.ANALYZE: 1>}

ExplainTypeLiteral: TypeAlias = Literal["analyze", "standard", "ANALYZE", "STANDARD"]

class PythonExceptionHandling(CppEnum):
    DEFAULT: ClassVar[PythonExceptionHandling]  # value = <PythonExceptionHandling.DEFAULT: 0>
    RETURN_NULL: ClassVar[PythonExceptionHandling]  # value = <PythonExceptionHandling.RETURN_NULL: 1>
    __members__: ClassVar[
        dict[str, PythonExceptionHandling]
    ]  # value = {'DEFAULT': <PythonExceptionHandling.DEFAULT: 0>, 'RETURN_NULL': <PythonExceptionHandling.RETURN_NULL: 1>}  # noqa: E501

class RenderMode(CppEnum):
    COLUMNS: ClassVar[RenderMode]  # value = <RenderMode.COLUMNS: 1>
    ROWS: ClassVar[RenderMode]  # value = <RenderMode.ROWS: 0>
    __members__: ClassVar[
        dict[str, RenderMode]
    ]  # value = {'ROWS': <RenderMode.ROWS: 0>, 'COLUMNS': <RenderMode.COLUMNS: 1>}

RenderModeLiteral: TypeAlias = Literal["columns", "rows", "COLUMNS", "ROWS"]

class StatementType(CppEnum):
    ALTER: ClassVar[StatementType]  # value = <StatementType.ALTER: 8>
    ANALYZE: ClassVar[StatementType]  # value = <StatementType.ANALYZE: 11>
    ATTACH: ClassVar[StatementType]  # value = <StatementType.ATTACH: 25>
    CALL: ClassVar[StatementType]  # value = <StatementType.CALL: 19>
    COPY_DATABASE: ClassVar[StatementType]  # value = <StatementType.COPY_DATABASE: 28>
    COPY: ClassVar[StatementType]  # value = <StatementType.COPY: 10>
    CREATE_FUNC: ClassVar[StatementType]  # value = <StatementType.CREATE_FUNC: 13>
    CREATE: ClassVar[StatementType]  # value = <StatementType.CREATE: 4>
    DELETE: ClassVar[StatementType]  # value = <StatementType.DELETE: 5>
    DETACH: ClassVar[StatementType]  # value = <StatementType.DETACH: 26>
    DROP: ClassVar[StatementType]  # value = <StatementType.DROP: 15>
    EXECUTE: ClassVar[StatementType]  # value = <StatementType.EXECUTE: 7>
    EXPLAIN: ClassVar[StatementType]  # value = <StatementType.EXPLAIN: 14>
    EXPORT: ClassVar[StatementType]  # value = <StatementType.EXPORT: 16>
    EXTENSION: ClassVar[StatementType]  # value = <StatementType.EXTENSION: 23>
    INSERT: ClassVar[StatementType]  # value = <StatementType.INSERT: 2>
    INVALID: ClassVar[StatementType]  # value = <StatementType.INVALID: 0>
    LOAD: ClassVar[StatementType]  # value = <StatementType.LOAD: 21>
    LOGICAL_PLAN: ClassVar[StatementType]  # value = <StatementType.LOGICAL_PLAN: 24>
    MERGE_INTO: ClassVar[StatementType]  # value = <StatementType.MERGE_INTO: 30>
    MULTI: ClassVar[StatementType]  # value = <StatementType.MULTI: 27>
    PRAGMA: ClassVar[StatementType]  # value = <StatementType.PRAGMA: 17>
    PREPARE: ClassVar[StatementType]  # value = <StatementType.PREPARE: 6>
    RELATION: ClassVar[StatementType]  # value = <StatementType.RELATION: 22>
    SELECT: ClassVar[StatementType]  # value = <StatementType.SELECT: 1>
    SET: ClassVar[StatementType]  # value = <StatementType.SET: 20>
    TRANSACTION: ClassVar[StatementType]  # value = <StatementType.TRANSACTION: 9>
    UPDATE: ClassVar[StatementType]  # value = <StatementType.UPDATE: 3>
    VACUUM: ClassVar[StatementType]  # value = <StatementType.VACUUM: 18>
    VARIABLE_SET: ClassVar[StatementType]  # value = <StatementType.VARIABLE_SET: 12>
    __members__: ClassVar[
        dict[str, StatementType]
    ]  # value = {'INVALID': <StatementType.INVALID: 0>, 'SELECT': <StatementType.SELECT: 1>, 'INSERT': <StatementType.INSERT: 2>, 'UPDATE': <StatementType.UPDATE: 3>, 'CREATE': <StatementType.CREATE: 4>, 'DELETE': <StatementType.DELETE: 5>, 'PREPARE': <StatementType.PREPARE: 6>, 'EXECUTE': <StatementType.EXECUTE: 7>, 'ALTER': <StatementType.ALTER: 8>, 'TRANSACTION': <StatementType.TRANSACTION: 9>, 'COPY': <StatementType.COPY: 10>, 'ANALYZE': <StatementType.ANALYZE: 11>, 'VARIABLE_SET': <StatementType.VARIABLE_SET: 12>, 'CREATE_FUNC': <StatementType.CREATE_FUNC: 13>, 'EXPLAIN': <StatementType.EXPLAIN: 14>, 'DROP': <StatementType.DROP: 15>, 'EXPORT': <StatementType.EXPORT: 16>, 'PRAGMA': <StatementType.PRAGMA: 17>, 'VACUUM': <StatementType.VACUUM: 18>, 'CALL': <StatementType.CALL: 19>, 'SET': <StatementType.SET: 20>, 'LOAD': <StatementType.LOAD: 21>, 'RELATION': <StatementType.RELATION: 22>, 'EXTENSION': <StatementType.EXTENSION: 23>, 'LOGICAL_PLAN': <StatementType.LOGICAL_PLAN: 24>, 'ATTACH': <StatementType.ATTACH: 25>, 'DETACH': <StatementType.DETACH: 26>, 'MULTI': <StatementType.MULTI: 27>, 'COPY_DATABASE': <StatementType.COPY_DATABASE: 28>, 'MERGE_INTO': <StatementType.MERGE_INTO: 30>}  # noqa: E501

class token_type(CppEnum):
    __members__: ClassVar[
        dict[str, token_type]
    ]  # value = {'identifier': <token_type.identifier: 0>, 'numeric_const': <token_type.numeric_const: 1>, 'string_const': <token_type.string_const: 2>, 'operator': <token_type.operator: 3>, 'keyword': <token_type.keyword: 4>, 'comment': <token_type.comment: 5>}  # noqa: E501
    comment: ClassVar[token_type]  # value = <token_type.comment: 5>
    identifier: ClassVar[token_type]  # value = <token_type.identifier: 0>
    keyword: ClassVar[token_type]  # value = <token_type.keyword: 4>
    numeric_const: ClassVar[token_type]  # value = <token_type.numeric_const: 1>
    operator: ClassVar[token_type]  # value = <token_type.operator: 3>
    string_const: ClassVar[token_type]  # value = <token_type.string_const: 2>



from typing import cast

from duckdb.sqltypes import DuckDBPyType

from ..exception import ContributionsAcceptedError
from .types import (
    ArrayType,
    BinaryType,
    BitstringType,
    BooleanType,
    ByteType,
    DataType,
    DateType,
    DayTimeIntervalType,
    DecimalType,
    DoubleType,
    FloatType,
    HugeIntegerType,
    IntegerType,
    LongType,
    MapType,
    ShortType,
    StringType,
    StructField,
    StructType,
    TimeNSType,
    TimeNTZType,
    TimestampMillisecondNTZType,
    TimestampNanosecondNTZType,
    TimestampNTZType,
    TimestampSecondNTZType,
    TimestampType,
    TimeType,
    UnsignedByteType,
    UnsignedHugeIntegerType,
    UnsignedIntegerType,
    UnsignedLongType,
    UnsignedShortType,
    UUIDType,
    VariantType,
)

_sqltype_to_spark_class = {
    "boolean": BooleanType,
    "utinyint": UnsignedByteType,
    "tinyint": ByteType,
    "usmallint": UnsignedShortType,
    "smallint": ShortType,
    "uinteger": UnsignedIntegerType,
    "integer": IntegerType,
    "ubigint": UnsignedLongType,
    "bigint": LongType,
    "hugeint": HugeIntegerType,
    "uhugeint": UnsignedHugeIntegerType,
    "varchar": StringType,
    "blob": BinaryType,
    "bit": BitstringType,
    "uuid": UUIDType,
    "date": DateType,
    "time": TimeNTZType,
    "time_ns": TimeNSType,
    "time with time zone": TimeType,
    "timestamp": TimestampNTZType,
    "timestamp with time zone": TimestampType,
    "timestamp_ms": TimestampNanosecondNTZType,
    "timestamp_ns": TimestampMillisecondNTZType,
    "timestamp_s": TimestampSecondNTZType,
    "interval": DayTimeIntervalType,
    "list": ArrayType,
    "struct": StructType,
    "map": MapType,
    # union
    # enum
    # null (???)
    "float": FloatType,
    "double": DoubleType,
    "decimal": DecimalType,
    "variant": VariantType,
}


def convert_nested_type(dtype: DuckDBPyType) -> DataType:  # noqa: D103
    id = dtype.id
    if id == "list" or id == "array":
        children = dtype.children
        return ArrayType(convert_type(children[0][1]))
    if id == "union":
        msg = (
            "Union types are not supported in the PySpark interface. "
            "DuckDB union types cannot be directly mapped to PySpark types."
        )
        raise ContributionsAcceptedError(msg)
    if id == "struct":
        children: list[tuple[str, DuckDBPyType]] = dtype.children
        fields = [StructField(x[0], convert_type(x[1])) for x in children]
        return StructType(fields)
    if id == "map":
        return MapType(convert_type(dtype.key), convert_type(dtype.value))
    raise NotImplementedError


def convert_type(dtype: DuckDBPyType) -> DataType:  # noqa: D103
    id = dtype.id
    if id in ["list", "struct", "map", "array"]:
        return convert_nested_type(dtype)
    if id == "decimal":
        children: list[tuple[str, DuckDBPyType]] = dtype.children
        precision = cast("int", children[0][1])
        scale = cast("int", children[1][1])
        return DecimalType(precision, scale)
    spark_type = _sqltype_to_spark_class[id]
    return spark_type()


def duckdb_to_spark_schema(names: list[str], types: list[DuckDBPyType]) -> StructType:  # noqa: D103
    fields = [StructField(name, dtype) for name, dtype in zip(names, [convert_type(x) for x in types], strict=False)]
    return StructType(fields)



from typing import ClassVar
from ._enums import CppEnum

__all__: list[str] = ["ARROW", "DEFAULT", "NATIVE", "SPECIAL", "FunctionNullHandling", "PythonUDFType"]

class FunctionNullHandling(CppEnum):
    DEFAULT: ClassVar[FunctionNullHandling]  # value = <FunctionNullHandling.DEFAULT: 0>
    SPECIAL: ClassVar[FunctionNullHandling]  # value = <FunctionNullHandling.SPECIAL: 1>
    __members__: ClassVar[
        dict[str, FunctionNullHandling]
    ]  # value = {'DEFAULT': <FunctionNullHandling.DEFAULT: 0>, 'SPECIAL': <FunctionNullHandling.SPECIAL: 1>}

class PythonUDFType(CppEnum):
    ARROW: ClassVar[PythonUDFType]  # value = <PythonUDFType.ARROW: 1>
    NATIVE: ClassVar[PythonUDFType]  # value = <PythonUDFType.NATIVE: 0>
    __members__: ClassVar[
        dict[str, PythonUDFType]
    ]  # value = {'NATIVE': <PythonUDFType.NATIVE: 0>, 'ARROW': <PythonUDFType.ARROW: 1>}

ARROW: PythonUDFType  # value = <PythonUDFType.ARROW: 1>
DEFAULT: FunctionNullHandling  # value = <FunctionNullHandling.DEFAULT: 0>
NATIVE: PythonUDFType  # value = <PythonUDFType.NATIVE: 0>
SPECIAL: FunctionNullHandling  # value = <FunctionNullHandling.SPECIAL: 1>



# ruff: noqa: D101, D104, D105, D107, ANN401
from typing import Any

from duckdb.sqltypes import (
    BIGINT,
    BIT,
    BLOB,
    BOOLEAN,
    DATE,
    DOUBLE,
    FLOAT,
    HUGEINT,
    INTEGER,
    INTERVAL,
    SMALLINT,
    SQLNULL,
    TIME,
    TIME_TZ,
    TIMESTAMP,
    TIMESTAMP_MS,
    TIMESTAMP_NS,
    TIMESTAMP_S,
    TIMESTAMP_TZ,
    TINYINT,
    UBIGINT,
    UHUGEINT,
    UINTEGER,
    USMALLINT,
    UTINYINT,
    UUID,
    VARCHAR,
    DuckDBPyType,
)


class Value:
    def __init__(self, object: Any, type: DuckDBPyType) -> None:
        self.object = object
        self.type = type

    def __repr__(self) -> str:
        return str(self.object)


# Miscellaneous


class NullValue(Value):
    def __init__(self) -> None:
        super().__init__(None, SQLNULL)


class BooleanValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, BOOLEAN)


# Unsigned numerics


class UnsignedBinaryValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, UTINYINT)


class UnsignedShortValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, USMALLINT)


class UnsignedIntegerValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, UINTEGER)


class UnsignedLongValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, UBIGINT)


# Signed numerics


class BinaryValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, TINYINT)


class ShortValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, SMALLINT)


class IntegerValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, INTEGER)


class LongValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, BIGINT)


class HugeIntegerValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, HUGEINT)


class UnsignedHugeIntegerValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, UHUGEINT)


# Fractional


class FloatValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, FLOAT)


class DoubleValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, DOUBLE)


class DecimalValue(Value):
    def __init__(self, object: Any, width: int, scale: int) -> None:
        import duckdb

        decimal_type = duckdb.decimal_type(width, scale)
        super().__init__(object, decimal_type)


# String


class StringValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, VARCHAR)


class UUIDValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, UUID)


class BitValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, BIT)


class BlobValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, BLOB)


# Temporal


class DateValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, DATE)


class IntervalValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, INTERVAL)


class TimestampValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, TIMESTAMP)


class TimestampSecondValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, TIMESTAMP_S)


class TimestampMillisecondValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, TIMESTAMP_MS)


class TimestampNanosecondValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, TIMESTAMP_NS)


class TimestampTimeZoneValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, TIMESTAMP_TZ)


class TimeValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, TIME)


class TimeTimeZoneValue(Value):
    def __init__(self, object: Any) -> None:
        super().__init__(object, TIME_TZ)


class ListValue(Value):
    def __init__(self, object: Any, child_type: DuckDBPyType) -> None:
        import duckdb

        list_type = duckdb.list_type(child_type)
        super().__init__(object, list_type)


class StructValue(Value):
    def __init__(self, object: Any, children: dict[str, DuckDBPyType]) -> None:
        import duckdb

        struct_type = duckdb.struct_type(children)
        super().__init__(object, struct_type)


class MapValue(Value):
    def __init__(self, object: Any, key_type: DuckDBPyType, value_type: DuckDBPyType) -> None:
        import duckdb

        map_type = duckdb.map_type(key_type, value_type)
        super().__init__(object, map_type)


class UnionType(Value):
    def __init__(self, object: Any, members: dict[str, DuckDBPyType]) -> None:
        import duckdb

        union_type = duckdb.union_type(members)
        super().__init__(object, union_type)


# TODO: add EnumValue once `duckdb.enum_type` is added  # noqa: TD002, TD003

__all__ = [
    "BinaryValue",
    "BitValue",
    "BlobValue",
    "BooleanValue",
    "DateValue",
    "DecimalValue",
    "DoubleValue",
    "FloatValue",
    "HugeIntegerValue",
    "IntegerValue",
    "IntervalValue",
    "LongValue",
    "NullValue",
    "ShortValue",
    "StringValue",
    "TimeTimeZoneValue",
    "TimeValue",
    "TimestampMillisecondValue",
    "TimestampNanosecondValue",
    "TimestampSecondValue",
    "TimestampTimeZoneValue",
    "TimestampValue",
    "UUIDValue",
    "UnsignedBinaryValue",
    "UnsignedHugeIntegerValue",
    "UnsignedIntegerValue",
    "UnsignedLongValue",
    "UnsignedShortValue",
    "Value",
]



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from collections.abc import Callable
from typing import (
    Any,
    TypeVar,
)

try:
    from typing import Literal, Protocol
except ImportError:
    from typing import Literal

    from typing_extensions import Protocol

import datetime
import decimal

from .._typing import PrimitiveType
from . import types
from .column import Column

ColumnOrName = Column | str
ColumnOrName_ = TypeVar("ColumnOrName_", bound=ColumnOrName)
DecimalLiteral = decimal.Decimal
DateTimeLiteral = datetime.datetime | datetime.date
LiteralType = PrimitiveType
AtomicDataTypeOrString = types.AtomicType | str
DataTypeOrString = types.DataType | str
OptionalPrimitiveType = PrimitiveType | None

AtomicValue = TypeVar(
    "AtomicValue",
    datetime.datetime,
    datetime.date,
    decimal.Decimal,
    bool,
    str,
    int,
    float,
)

RowLike = TypeVar("RowLike", list[Any], tuple[Any, ...], types.Row)

SQLBatchedUDFType = Literal[100]


class SupportsOpen(Protocol):
    def open(self, partition_id: int, epoch_id: int) -> bool: ...


class SupportsProcess(Protocol):
    def process(self, row: types.Row) -> None: ...


class SupportsClose(Protocol):
    def close(self, error: Exception) -> None: ...


class UserDefinedFunctionLike(Protocol):
    func: Callable[..., Any]
    evalType: int
    deterministic: bool

    @property
    def returnType(self) -> types.DataType: ...

    def __call__(self, *args: ColumnOrName) -> Column: ...

    def asNondeterministic(self) -> "UserDefinedFunctionLike": ...



from __future__ import annotations  # noqa: D100

import contextlib
import datetime
import json
import typing
from decimal import Decimal

import polars as pl
from polars.io.plugins import register_io_source

import duckdb

if typing.TYPE_CHECKING:
    from collections.abc import Iterator

    import typing_extensions

_ExpressionTree: typing_extensions.TypeAlias = typing.Dict[str, typing.Union[str, int, "_ExpressionTree", typing.Any]]  # noqa: UP006


def _predicate_to_expression(predicate: pl.Expr) -> duckdb.Expression | None:
    """Convert a Polars predicate expression to a DuckDB-compatible SQL expression.

    Parameters:
        predicate (pl.Expr): A Polars expression (e.g., col("foo") > 5)

    Returns:
        SQLExpression: A DuckDB SQL expression string equivalent.
        None: If conversion fails.

    Example:
        >>> _predicate_to_expression(pl.col("foo") > 5)
        SQLExpression("(foo > 5)")
    """
    # Serialize the Polars expression tree to JSON
    tree = json.loads(predicate.meta.serialize(format="json"))

    try:
        # Convert the tree to SQL
        sql_filter = _pl_tree_to_sql(tree)
        return duckdb.SQLExpression(sql_filter)
    except Exception:
        # If the conversion fails, we return None
        return None


def _pl_operation_to_sql(op: str) -> str:
    """Map Polars binary operation strings to SQL equivalents.

    Example:
        >>> _pl_operation_to_sql("Eq")
        '='
    """
    try:
        return {
            "Lt": "<",
            "LtEq": "<=",
            "Gt": ">",
            "GtEq": ">=",
            "Eq": "=",
            "Modulus": "%",
            "And": "AND",
            "Or": "OR",
        }[op]
    except KeyError:
        raise NotImplementedError(op)  # noqa: B904


def _escape_sql_identifier(identifier: str) -> str:
    """Escape SQL identifiers by doubling any double quotes and wrapping in double quotes.

    Example:
        >>> _escape_sql_identifier('column"name')
        '"column""name"'
    """
    escaped = identifier.replace('"', '""')
    return f'"{escaped}"'


def _pl_tree_to_sql(tree: _ExpressionTree) -> str:
    """Recursively convert a Polars expression tree (as JSON) to a SQL string.

    Parameters:
        tree (dict): JSON-deserialized expression tree from Polars

    Returns:
        str: SQL expression string

    Example:
        Input tree:
        {
            "BinaryExpr": {
                "left": { "Column": "foo" },
                "op": "Gt",
                "right": { "Literal": { "Int": 5 } }
            }
        }
        Output: "(foo > 5)"
    """
    [node_type] = tree.keys()

    if node_type == "BinaryExpr":
        # Binary expressions: left OP right
        bin_expr_tree = tree[node_type]
        assert isinstance(bin_expr_tree, dict), f"A {node_type} should be a dict but got {type(bin_expr_tree)}"
        lhs, op, rhs = bin_expr_tree["left"], bin_expr_tree["op"], bin_expr_tree["right"]
        assert isinstance(lhs, dict), f"LHS of a {node_type} should be a dict but got {type(lhs)}"
        assert isinstance(op, str), f"The op of a {node_type} should be a str but got {type(op)}"
        assert isinstance(rhs, dict), f"RHS of a {node_type} should be a dict but got {type(rhs)}"
        return f"({_pl_tree_to_sql(lhs)} {_pl_operation_to_sql(op)} {_pl_tree_to_sql(rhs)})"
    if node_type == "Column":
        # A reference to a column name
        # Wrap in quotes to handle special characters
        col_name = tree[node_type]
        assert isinstance(col_name, str), f"The col name of a {node_type} should be a str but got {type(col_name)}"
        return _escape_sql_identifier(col_name)

    if node_type in ("Literal", "Dyn"):
        # Recursively process dynamic or literal values
        val_tree = tree[node_type]
        assert isinstance(val_tree, dict), f"A {node_type} should be a dict but got {type(val_tree)}"
        return _pl_tree_to_sql(val_tree)

    if node_type == "Int":
        # Direct integer literals
        int_literal = tree[node_type]
        assert isinstance(int_literal, (int, str)), (
            f"The value of an Int should be an int or str but got {type(int_literal)}"
        )
        return str(int_literal)

    if node_type == "Float":
        # Direct float literals
        float_literal = tree[node_type]
        assert isinstance(float_literal, (float, int, str)), (
            f"The value of a Float should be a float, int or str but got {type(float_literal)}"
        )
        return str(float_literal)

    if node_type == "Function":
        # Handle boolean functions like IsNull, IsNotNull
        func_tree = tree[node_type]
        assert isinstance(func_tree, dict), f"A {node_type} should be a dict but got {type(func_tree)}"
        inputs = func_tree["input"]
        assert isinstance(inputs, list), f"A {node_type} should have a list of dicts as input but got {type(inputs)}"
        input_tree = inputs[0]
        assert isinstance(input_tree, dict), (
            f"A {node_type} should have a list of dicts as input but got {type(input_tree)}"
        )
        func_dict = func_tree["function"]
        assert isinstance(func_dict, dict), (
            f"A {node_type} should have a function dict as input but got {type(func_dict)}"
        )

        if "Boolean" in func_dict:
            func = func_dict["Boolean"]
            arg_sql = _pl_tree_to_sql(inputs[0])

            if func == "IsNull":
                return f"({arg_sql} IS NULL)"
            if func == "IsNotNull":
                return f"({arg_sql} IS NOT NULL)"
            msg = f"Boolean function not supported: {func}"
            raise NotImplementedError(msg)

        msg = f"Unsupported function type: {func_dict}"
        raise NotImplementedError(msg)

    if node_type == "Cast":
        cast_tree = tree[node_type]
        assert isinstance(cast_tree, dict), f"A {node_type} should be a dict but got {type(cast_tree)}"
        options = cast_tree.get("options")
        if options == "Strict":
            # Strict casts on literals (e.g. pl.lit(1, dtype=pl.Int8)) are safe to unwrap —
            # the value is known at expression creation time. Strict casts on columns
            # (e.g. pl.col("a").cast(pl.Int64)) are semantically meaningful and must not be dropped.
            cast_expr = cast_tree.get("expr", {})
            if not isinstance(cast_expr, dict) or "Literal" not in cast_expr:
                msg = "Strict cast on non-literal expression cannot be pushed down"
                raise NotImplementedError(msg)
        elif options != "NonStrict":
            msg = f"Only NonStrict/Strict casts can be safely unwrapped, got {options!r}"
            raise NotImplementedError(msg)
        cast_expr = cast_tree["expr"]
        assert isinstance(cast_expr, dict), f"A {node_type} should be a dict but got {type(cast_expr)}"
        return _pl_tree_to_sql(cast_expr)

    if node_type == "Scalar":
        # Detect format: old style (dtype/value) or new style (direct type key)
        scalar_tree = tree[node_type]
        assert isinstance(scalar_tree, dict), f"A {node_type} should be a dict but got {type(scalar_tree)}"
        if "dtype" in scalar_tree and "value" in scalar_tree:
            dtype = str(scalar_tree["dtype"])
            value = scalar_tree["value"]
        else:
            # New style: dtype is the single key in the dict
            dtype = next(iter(scalar_tree.keys()))
            value = scalar_tree
        assert isinstance(dtype, str), f"A {node_type} should have a str dtype but got  {type(dtype)}"
        assert isinstance(value, dict), f"A {node_type} should have a dict value but got {type(value)}"

        # Decimal support
        if dtype.startswith("{'Decimal'") or dtype == "Decimal":
            decimal_value = value["Decimal"]
            assert isinstance(decimal_value, list), (
                f"A {dtype} should be a two or three member list but got {type(decimal_value)}"
            )
            assert 2 <= len(decimal_value) <= 3, (
                f"A {dtype} should be a two or three member list but got {len(decimal_value)} member list"
            )
            return str(Decimal(decimal_value[0]) / Decimal(10 ** decimal_value[-1]))

        # Datetime with microseconds since epoch
        if dtype.startswith("{'Datetime'") or dtype == "Datetime":
            micros = value["Datetime"]
            assert isinstance(micros, list), f"A {dtype} should be a one member list but got {type(micros)}"
            dt_timestamp = datetime.datetime.fromtimestamp(micros[0] / 1_000_000, tz=datetime.timezone.utc)
            return f"'{dt_timestamp!s}'::TIMESTAMP"

        # Match simple numeric/boolean types
        if dtype in (
            "Int8",
            "Int16",
            "Int32",
            "Int64",
            "Int128",
            "UInt8",
            "UInt16",
            "UInt32",
            "UInt64",
            "UInt128",
            "Float32",
            "Float64",
            "Boolean",
        ):
            return str(value[dtype])

        # Time type
        if dtype == "Time":
            nanoseconds = value["Time"]
            assert isinstance(nanoseconds, int), f"A {dtype} should be an int but got {type(nanoseconds)}"
            seconds = nanoseconds // 1_000_000_000
            microseconds = (nanoseconds % 1_000_000_000) // 1_000
            dt_time = (datetime.datetime.min + datetime.timedelta(seconds=seconds, microseconds=microseconds)).time()
            return f"'{dt_time}'::TIME"

        # Date type
        if dtype == "Date":
            days_since_epoch = value["Date"]
            assert isinstance(days_since_epoch, (float, int)), (
                f"A {dtype} should be a number but got {type(days_since_epoch)}"
            )
            date = datetime.date(1970, 1, 1) + datetime.timedelta(days=days_since_epoch)
            return f"'{date}'::DATE"

        # Binary type
        if dtype == "Binary":
            bin_value = value["Binary"]
            assert isinstance(bin_value, list), f"A {dtype} should be a list but got {type(bin_value)}"
            binary_data = bytes(bin_value)
            escaped = "".join(f"\\x{b:02x}" for b in binary_data)
            return f"'{escaped}'::BLOB"

        # String type
        if dtype == "String" or dtype == "StringOwned":
            # Some new formats may store directly under StringOwned
            string_val = value.get("StringOwned", value.get("String", None))
            # the string must be a string constant
            return str(duckdb.ConstantExpression(string_val))

        msg = f"Unsupported scalar type {dtype!s}, with value {value}"
        raise NotImplementedError(msg)

    msg = f"Node type: {node_type} is not implemented. {tree[node_type]}"
    raise NotImplementedError(msg)


def duckdb_source(relation: duckdb.DuckDBPyRelation, schema: pl.schema.Schema) -> pl.LazyFrame:
    """A polars IO plugin for DuckDB."""

    def source_generator(
        with_columns: list[str] | None,
        predicate: pl.Expr | None,
        n_rows: int | None,
        batch_size: int | None,
    ) -> Iterator[pl.DataFrame]:
        duck_predicate = None
        relation_final = relation
        if with_columns is not None:
            cols = ",".join(map(_escape_sql_identifier, with_columns))
            relation_final = relation_final.project(cols)
        if n_rows is not None:
            relation_final = relation_final.limit(n_rows)
        if predicate is not None:
            # We have a predicate, if possible, we push it down to DuckDB
            with contextlib.suppress(AssertionError, KeyError):
                duck_predicate = _predicate_to_expression(predicate)
        # Try to pushdown filter, if one exists
        if duck_predicate is not None:
            relation_final = relation_final.filter(duck_predicate)
        results = relation_final.to_arrow_reader() if batch_size is None else relation_final.to_arrow_reader(batch_size)

        for record_batch in iter(results.read_next_batch, None):
            if predicate is not None and duck_predicate is None:
                # We have a predicate, but did not manage to push it down, we fallback here
                yield pl.from_arrow(record_batch).filter(predicate)  # type: ignore[arg-type,misc,unused-ignore]
            else:
                yield pl.from_arrow(record_batch)  # type: ignore[misc,unused-ignore]

    return register_io_source(source_generator, schema=schema, is_pure=True)



# https://sparkbyexamples.com/pyspark/pyspark-udf-user-defined-function/
from typing import TYPE_CHECKING, Any, Optional, TypeVar

if TYPE_CHECKING:
    from collections.abc import Callable

from .types import DataType

if TYPE_CHECKING:
    from .session import SparkSession

DataTypeOrString = DataType | str
UserDefinedFunctionLike = TypeVar("UserDefinedFunctionLike")


class UDFRegistration:  # noqa: D101
    def __init__(self, sparkSession: "SparkSession") -> None:  # noqa: D107
        self.sparkSession = sparkSession

    def register(  # noqa: D102
        self,
        name: str,
        f: "Callable[..., Any] | UserDefinedFunctionLike",
        returnType: Optional["DataTypeOrString"] = None,
    ) -> "UserDefinedFunctionLike":
        self.sparkSession.conn.create_function(name, f, return_type=returnType)

    def registerJavaFunction(  # noqa: D102
        self,
        name: str,
        javaClassName: str,
        returnType: Optional["DataTypeOrString"] = None,
    ) -> None:
        raise NotImplementedError

    def registerJavaUDAF(self, name: str, javaClassName: str) -> None:  # noqa: D102
        raise NotImplementedError


__all__ = ["UDFRegistration"]



# ----------------------------------------------------------------------
# Version API
#
# We provide three symbols:
# - duckdb.__version__: The version of this package
# - duckdb.__duckdb_version__: The version of duckdb that is bundled
# - duckdb.version(): A human-readable version string containing both of the above
# ----------------------------------------------------------------------
from importlib.metadata import version as _dist_version

import _duckdb

__version__: str = _dist_version("duckdb")
"""Version of the DuckDB Python Package."""

__duckdb_version__: str = _duckdb.__version__
"""Version of DuckDB that is bundled."""


def version() -> str:
    """Human-friendly formatted version string of both the distribution package and the bundled DuckDB engine."""
    return f"{__version__} (with duckdb {_duckdb.__version__})"



from .conf import SparkConf  # noqa: D104
from .context import SparkContext
from .exception import ContributionsAcceptedError
from .sql import DataFrame, SparkSession

__all__ = ["ContributionsAcceptedError", "DataFrame", "SparkConf", "SparkContext", "SparkSession"]



from typing import cast

from ..utils import ErrorClassesReader


class PySparkException(Exception):
    """Base Exception for handling errors generated from PySpark."""

    def __init__(  # noqa: D107
        self,
        message: str | None = None,
        # The error class, decides the message format, must be one of the valid options listed in 'error_classes.py'
        error_class: str | None = None,
        # The dictionary listing the arguments specified in the message (or the error_class)
        message_parameters: dict[str, str] | None = None,
    ) -> None:
        # `message` vs `error_class` & `message_parameters` are mutually exclusive.
        assert (message is not None and (error_class is None and message_parameters is None)) or (
            message is None and (error_class is not None and message_parameters is not None)
        )

        self.error_reader = ErrorClassesReader()

        if message is None:
            self.message = self.error_reader.get_error_message(
                cast("str", error_class), cast("dict[str, str]", message_parameters)
            )
        else:
            self.message = message

        self.error_class = error_class
        self.message_parameters = message_parameters

    def getErrorClass(self) -> str | None:
        """Returns an error class as a string.

        .. versionadded:: 3.4.0

        See Also:
        --------
        :meth:`PySparkException.getMessageParameters`
        :meth:`PySparkException.getSqlState`
        """
        return self.error_class

    def getMessageParameters(self) -> dict[str, str] | None:
        """Returns a message parameters as a dictionary.

        .. versionadded:: 3.4.0

        See Also:
        --------
        :meth:`PySparkException.getErrorClass`
        :meth:`PySparkException.getSqlState`
        """
        return self.message_parameters

    def getSqlState(self) -> None:
        """Returns an SQLSTATE as a string.

        Errors generated in Python have no SQLSTATE, so it always returns None.

        .. versionadded:: 3.4.0

        See Also:
        --------
        :meth:`PySparkException.getErrorClass`
        :meth:`PySparkException.getMessageParameters`
        """
        return None

    def __str__(self) -> str:  # noqa: D105
        if self.getErrorClass() is not None:
            return f"[{self.getErrorClass()}] {self.message}"
        else:
            return self.message


class AnalysisException(PySparkException):
    """Failed to analyze a SQL query plan."""


class SessionNotSameException(PySparkException):
    """Performed the same operation on different SparkSession."""


class TempTableAlreadyExistsException(AnalysisException):
    """Failed to create temp view since it is already exists."""


class ParseException(AnalysisException):
    """Failed to parse a SQL command."""


class IllegalArgumentException(PySparkException):
    """Passed an illegal or inappropriate argument."""


class ArithmeticException(PySparkException):
    """Arithmetic exception thrown from Spark with an error class."""


class UnsupportedOperationException(PySparkException):
    """Unsupported operation exception thrown from Spark with an error class."""


class ArrayIndexOutOfBoundsException(PySparkException):
    """Array index out of bounds exception thrown from Spark with an error class."""


class DateTimeException(PySparkException):
    """Datetime exception thrown from Spark with an error class."""


class NumberFormatException(IllegalArgumentException):
    """Number format exception thrown from Spark with an error class."""


class StreamingQueryException(PySparkException):
    """Exception that stopped a :class:`StreamingQuery`."""


class QueryExecutionException(PySparkException):
    """Failed to execute a query."""


class PythonException(PySparkException):
    """Exceptions thrown from Python workers."""


class SparkRuntimeException(PySparkException):
    """Runtime exception thrown from Spark with an error class."""


class SparkUpgradeException(PySparkException):
    """Exception thrown because of Spark upgrade."""


class UnknownException(PySparkException):
    """None of the above exceptions."""


class PySparkValueError(PySparkException, ValueError):
    """Wrapper class for ValueError to support error classes."""


class PySparkIndexError(PySparkException, IndexError):
    """Wrapper class for IndexError to support error classes."""


class PySparkTypeError(PySparkException, TypeError):
    """Wrapper class for TypeError to support error classes."""


class PySparkAttributeError(PySparkException, AttributeError):
    """Wrapper class for AttributeError to support error classes."""


class PySparkRuntimeError(PySparkException, RuntimeError):
    """Wrapper class for RuntimeError to support error classes."""


class PySparkAssertionError(PySparkException, AssertionError):
    """Wrapper class for AssertionError to support error classes."""


class PySparkNotImplementedError(PySparkException, NotImplementedError):
    """Wrapper class for NotImplementedError to support error classes."""



import uuid
from collections.abc import Iterable, Sized
from typing import TYPE_CHECKING, Any, NoReturn, Union

import duckdb

if TYPE_CHECKING:
    from pandas.core.frame import DataFrame as PandasDataFrame

    from .catalog import Catalog


from ..conf import SparkConf
from ..context import SparkContext
from ..errors import PySparkTypeError
from ..exception import ContributionsAcceptedError
from .conf import RuntimeConfig
from .dataframe import DataFrame
from .readwriter import DataFrameReader
from .streaming import DataStreamReader
from .types import StructType
from .udf import UDFRegistration

# In spark:
# SparkSession holds a SparkContext
# SparkContext gets created from SparkConf
# At this level the check is made to determine whether the instance already exists and just needs
# to be retrieved or it needs to be created.

# For us this is done inside of `duckdb.connect`, based on the passed in path + configuration
# SparkContext can be compared to our Connection class, and SparkConf to our ClientContext class


# data is a List of rows
# every value in each row needs to be turned into a Value
def _combine_data_and_schema(data: Iterable[Any], schema: StructType) -> list[duckdb.Value]:
    from duckdb import Value

    new_data = []
    for row in data:
        new_row = [Value(x, dtype.duckdb_type) for x, dtype in zip(row, [y.dataType for y in schema], strict=False)]
        new_data.append(new_row)
    return new_data


class SparkSession:  # noqa: D101
    def __init__(self, context: SparkContext) -> None:  # noqa: D107
        self.conn = context.connection
        self._context = context
        self._conf = RuntimeConfig(self.conn)

    def _create_dataframe(self, data: Union[Iterable[Any], "PandasDataFrame"]) -> DataFrame:
        try:
            import pandas

            has_pandas = True
        except ImportError:
            has_pandas = False
        if has_pandas and isinstance(data, pandas.DataFrame):
            unique_name = f"pyspark_pandas_df_{uuid.uuid1()}"
            self.conn.register(unique_name, data)
            return DataFrame(self.conn.sql(f'select * from "{unique_name}"'), self)

        def verify_tuple_integrity(tuples: list[tuple]) -> None:
            if len(tuples) <= 1:
                return
            expected_length = len(tuples[0])
            for i, item in enumerate(tuples[1:]):
                actual_length = len(item)
                if expected_length == actual_length:
                    continue
                raise PySparkTypeError(
                    error_class="LENGTH_SHOULD_BE_THE_SAME",
                    message_parameters={
                        "arg1": f"data{i}",
                        "arg2": f"data{i + 1}",
                        "arg1_length": str(expected_length),
                        "arg2_length": str(actual_length),
                    },
                )

        if not isinstance(data, list):
            data = list(data)
        verify_tuple_integrity(data)

        def construct_query(tuples: Iterable) -> str:
            def construct_values_list(row: Sized, start_param_idx: int) -> str:
                parameter_count = len(row)
                parameters = [f"${x + start_param_idx}" for x in range(parameter_count)]
                parameters = "(" + ", ".join(parameters) + ")"
                return parameters

            row_size = len(tuples[0])
            values_list = [construct_values_list(x, 1 + (i * row_size)) for i, x in enumerate(tuples)]
            values_list = ", ".join(values_list)

            query = f"""
                select * from (values {values_list})
            """
            return query

        query = construct_query(data)

        def construct_parameters(tuples: Iterable) -> list[list]:
            parameters = []
            for row in tuples:
                parameters.extend(list(row))
            return parameters

        parameters = construct_parameters(data)

        rel = self.conn.sql(query, params=parameters)
        return DataFrame(rel, self)

    def _createDataFrameFromPandas(
        self, data: "PandasDataFrame", types: list[str] | None, names: list[str] | None
    ) -> DataFrame:
        df = self._create_dataframe(data)

        # Cast to types
        if types:
            df = df._cast_types(*types)
        # Alias to names
        if names:
            df = df.toDF(*names)
        return df

    def createDataFrame(  # noqa: D102
        self,
        data: Union["PandasDataFrame", Iterable[Any]],
        schema: StructType | list[str] | None = None,
        samplingRatio: float | None = None,
        verifySchema: bool = True,
    ) -> DataFrame:
        if samplingRatio:
            raise NotImplementedError
        if not verifySchema:
            raise NotImplementedError
        types = None
        names = None

        if isinstance(data, DataFrame):
            raise PySparkTypeError(
                error_class="SHOULD_NOT_DATAFRAME",
                message_parameters={"arg_name": "data"},
            )

        if schema:
            if isinstance(schema, StructType):
                types, names = schema.extract_types_and_names()
            else:
                names = schema

        try:
            import pandas

            has_pandas = True
        except ImportError:
            has_pandas = False
        # Falsey check on pandas dataframe is not defined, so first check if it's not a pandas dataframe
        # Then check if 'data' is None or []
        if has_pandas and isinstance(data, pandas.DataFrame):
            return self._createDataFrameFromPandas(data, types, names)

        # Finally check if a schema was provided
        is_empty = False
        if not data and names:
            # Create NULLs for every type in our dataframe
            is_empty = True
            data = [tuple(None for _ in names)]

        if schema and isinstance(schema, StructType):
            # Transform the data into Values to combine the data+schema
            data = _combine_data_and_schema(data, schema)

        df = self._create_dataframe(data)
        if is_empty:
            rel = df.relation
            # Add impossible where clause
            rel = rel.filter("1=0")
            df = DataFrame(rel, self)

        # Cast to types
        if types:
            df = df._cast_types(*types)
        # Alias to names
        if names:
            df = df.toDF(*names)
        return df

    def newSession(self) -> "SparkSession":  # noqa: D102
        return SparkSession(self._context)

    def range(  # noqa: D102
        self,
        start: int,
        end: int | None = None,
        step: int = 1,
        numPartitions: int | None = None,
    ) -> "DataFrame":
        if numPartitions:
            raise ContributionsAcceptedError

        if end is None:
            end = start
            start = 0

        return DataFrame(self.conn.table_function("range", parameters=[start, end, step]), self)

    def sql(self, sqlQuery: str, **kwargs: Any) -> DataFrame:  # noqa: D102, ANN401
        if kwargs:
            raise NotImplementedError
        relation = self.conn.sql(sqlQuery)
        return DataFrame(relation, self)

    def stop(self) -> None:  # noqa: D102
        self._context.stop()

    def table(self, tableName: str) -> DataFrame:  # noqa: D102
        relation = self.conn.table(tableName)
        return DataFrame(relation, self)

    def getActiveSession(self) -> "SparkSession":  # noqa: D102
        return self

    @property
    def catalog(self) -> "Catalog":  # noqa: D102
        if not hasattr(self, "_catalog"):
            from duckdb.experimental.spark.sql.catalog import Catalog

            self._catalog = Catalog(self)
        return self._catalog

    @property
    def conf(self) -> RuntimeConfig:  # noqa: D102
        return self._conf

    @property
    def read(self) -> DataFrameReader:  # noqa: D102
        return DataFrameReader(self)

    @property
    def readStream(self) -> DataStreamReader:  # noqa: D102
        return DataStreamReader(self)

    @property
    def sparkContext(self) -> SparkContext:  # noqa: D102
        return self._context

    @property
    def streams(self) -> NoReturn:  # noqa: D102
        raise ContributionsAcceptedError

    @property
    def udf(self) -> UDFRegistration:  # noqa: D102
        return UDFRegistration(self)

    @property
    def version(self) -> str:  # noqa: D102
        return "1.0.0"

    class Builder:  # noqa: D106
        def __init__(self) -> None:  # noqa: D107
            pass

        def master(self, name: str) -> "SparkSession.Builder":  # noqa: D102
            # no-op
            return self

        def appName(self, name: str) -> "SparkSession.Builder":  # noqa: D102
            # no-op
            return self

        def remote(self, url: str) -> "SparkSession.Builder":  # noqa: D102
            # no-op
            return self

        def getOrCreate(self) -> "SparkSession":  # noqa: D102
            context = SparkContext("__ignored__")
            return SparkSession(context)

        def config(  # noqa: D102
            self,
            key: str | None = None,
            value: Any | None = None,  # noqa: ANN401
            conf: SparkConf | None = None,
        ) -> "SparkSession.Builder":
            return self

        def enableHiveSupport(self) -> "SparkSession.Builder":  # noqa: D102
            # no-op
            return self

    builder = Builder()


__all__ = ["SparkSession"]
